#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
module: nxos_vxlan_vtep_vni
extends_documentation_fragment:
- cisco.nxos.nxos
short_description: Creates a Virtual Network Identifier member (VNI)
description:
- Creates a Virtual Network Identifier member (VNI) for an NVE overlay interface.
version_added: 1.0.0
author: Gabriele Gerbino (@GGabriele)
notes:
- Tested against NXOSv 7.3.(0)D1(1) on VIRL
- Unsupported for Cisco MDS
- default, where supported, restores params default value.
options:
  interface:
    description:
    - Interface name for the VXLAN Network Virtualization Endpoint.
    required: true
    type: str
  vni:
    description:
    - ID of the Virtual Network Identifier.
    required: true
    type: str
  assoc_vrf:
    description:
    - This attribute is used to identify and separate processing VNIs that are associated
      with a VRF and used for routing. The VRF and VNI specified with this command
      must match the configuration of the VNI under the VRF.
    type: bool
  ingress_replication:
    description:
    - Specifies mechanism for host reachability advertisement.
    choices:
    - bgp
    - static
    - default
    type: str
  multicast_group:
    description:
    - The multicast group (range) of the VNI. Valid values are string and keyword
      'default'.
    type: str
  peer_list:
    description:
    - Set the ingress-replication static peer list. Valid values are an array, a space-separated
      string of ip addresses, or the keyword 'default'.
    type: list
    elements: str
  suppress_arp:
    description:
    - Suppress arp under layer 2 VNI.
    type: bool
  suppress_arp_disable:
    description:
    - Overrides the global ARP suppression config. This is available on NX-OS 9K series
      running 9.2.x or higher.
    type: bool
  state:
    description:
    - Determines whether the config should be present or not on the device.
    default: present
    choices:
    - present
    - absent
    type: str
  multisite_ingress_replication:
    description:
    - Enables multisite ingress replication.
    choices:
    - disable
    - enable
    - optimized
    type: str
    version_added: 1.1.0
"""
EXAMPLES = """
- cisco.nxos.nxos_vxlan_vtep_vni:
    interface: nve1
    vni: 6000
    ingress_replication: default
    multisite_ingress_replication: enable
"""

RETURN = """
commands:
    description: commands sent to the device
    returned: always
    type: list
    sample: ["interface nve1", "member vni 6000", "multisite ingress-replication"]
"""

import re

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.config import (
    CustomNetworkConfig,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    get_config,
    load_config,
)


BOOL_PARAMS = ["assoc_vrf", "suppress_arp", "suppress_arp_disable"]
PARAM_TO_DEFAULT_KEYMAP = {
    "multicast_group": "",
    "peer_list": [],
    "ingress_replication": "",
}
PARAM_TO_COMMAND_KEYMAP = {
    "assoc_vrf": "associate-vrf",
    "interface": "interface",
    "vni": "member vni",
    "ingress_replication": "ingress-replication protocol",
    "multicast_group": "mcast-group",
    "peer_list": "peer-ip",
    "suppress_arp": "suppress-arp",
    "suppress_arp_disable": "suppress-arp disable",
    "multisite_ingress_replication": "multisite ingress-replication",
}


def get_value(arg, config, module):
    command = PARAM_TO_COMMAND_KEYMAP[arg]
    command_val_re = re.compile(r"(?:{0}\s)(?P<value>.*)$".format(command), re.M)

    if arg in BOOL_PARAMS:
        command_re = re.compile(r"\s+{0}\s*$".format(command), re.M)
        value = False
        if command_re.search(config):
            value = True
    elif arg == "peer_list":
        has_command_val = command_val_re.findall(config, re.M)
        value = []
        if has_command_val:
            value = has_command_val
    elif arg == "multisite_ingress_replication":
        has_command = re.search(r"^\s+{0}$".format(command), config, re.M)
        has_command_val = command_val_re.search(config, re.M)
        value = "disable"
        if has_command:
            value = "enable"
        elif has_command_val:
            value = "optimized"
    else:
        value = ""
        has_command_val = command_val_re.search(config, re.M)
        if has_command_val:
            value = has_command_val.group("value")
    return value


def check_interface(module, netcfg):
    config = str(netcfg)

    has_interface = re.search(r"(?:interface nve)(?P<value>.*)$", config, re.M)
    value = ""
    if has_interface:
        value = "nve{0}".format(has_interface.group("value"))

    return value


def get_existing(module, args):
    existing = {}
    netcfg = CustomNetworkConfig(indent=2, contents=get_config(module))

    interface_exist = check_interface(module, netcfg)
    if interface_exist:
        parents = ["interface {0}".format(interface_exist)]
        temp_config = netcfg.get_section(parents)

        if "member vni {0} associate-vrf".format(module.params["vni"]) in temp_config:
            parents.append("member vni {0} associate-vrf".format(module.params["vni"]))
            config = netcfg.get_section(parents)
        elif "member vni {0}".format(module.params["vni"]) in temp_config:
            parents.append("member vni {0}".format(module.params["vni"]))
            config = netcfg.get_section(parents)
        else:
            config = {}

        if config:
            for arg in args:
                if arg not in ["interface", "vni"]:
                    existing[arg] = get_value(arg, config, module)
            existing["interface"] = interface_exist
            existing["vni"] = module.params["vni"]

    return existing, interface_exist


def apply_key_map(key_map, table):
    new_dict = {}
    for key, value in table.items():
        new_key = key_map.get(key)
        if new_key:
            new_dict[new_key] = value
    return new_dict


def state_present(module, existing, proposed, candidate):
    commands = list()
    proposed_commands = apply_key_map(PARAM_TO_COMMAND_KEYMAP, proposed)
    existing_commands = apply_key_map(PARAM_TO_COMMAND_KEYMAP, existing)

    for key, value in proposed_commands.items():
        if key == "associate-vrf":
            command = "member vni {0} {1}".format(module.params["vni"], key)
            if not value:
                command = "no {0}".format(command)
            commands.append(command)

        elif key == "peer-ip" and value != []:
            for peer in value:
                commands.append("{0} {1}".format(key, peer))

        elif key == "mcast-group" and value != existing_commands.get(key):
            commands.append("no {0}".format(key))
            vni_command = "member vni {0}".format(module.params["vni"])
            if vni_command not in commands:
                commands.append("member vni {0}".format(module.params["vni"]))
            if value != PARAM_TO_DEFAULT_KEYMAP.get("multicast_group", "default"):
                commands.append("{0} {1}".format(key, value))

        elif key == "ingress-replication protocol" and value != existing_commands.get(key):
            evalue = existing_commands.get(key)
            dvalue = PARAM_TO_DEFAULT_KEYMAP.get("ingress_replication", "default")
            if value != dvalue:
                if evalue and evalue != dvalue:
                    commands.append("no {0} {1}".format(key, evalue))
                commands.append("{0} {1}".format(key, value))
            else:
                if evalue:
                    commands.append("no {0} {1}".format(key, evalue))

        elif value is True:
            commands.append(key)
        elif value is False:
            commands.append("no {0}".format(key))
        elif value == "default" or value == []:
            if existing_commands.get(key):
                existing_value = existing_commands.get(key)
                if key == "peer-ip":
                    for peer in existing_value:
                        commands.append("no {0} {1}".format(key, peer))
                else:
                    commands.append("no {0} {1}".format(key, existing_value))
            else:
                if key.replace(" ", "_").replace("-", "_") in BOOL_PARAMS:
                    commands.append("no {0}".format(key.lower()))
        elif key == "multisite ingress-replication" and value != existing_commands.get(key):
            vni_command = "member vni {0}".format(module.params["vni"])
            if vni_command not in commands:
                commands.append("member vni {0}".format(module.params["vni"]))
            if value == "disable":
                command = "no {0}".format(key)
                commands.append(command)
            elif value == "enable":
                command = "{0}".format(key)
                commands.append(command)
            elif value == "optimized":
                command = "{0} {1}".format(key, value)
                commands.append(command)
        else:
            command = "{0} {1}".format(key, value.lower())
            commands.append(command)

    if commands:
        vni_command = "member vni {0}".format(module.params["vni"])
        ingress_replications_command = "ingress-replication protocol static"
        ingress_replicationb_command = "ingress-replication protocol bgp"
        ingress_replicationns_command = "no ingress-replication protocol static"
        ingress_replicationnb_command = "no ingress-replication protocol bgp"
        interface_command = "interface {0}".format(module.params["interface"])

        if any(
            c in commands
            for c in (
                ingress_replications_command,
                ingress_replicationb_command,
                ingress_replicationnb_command,
                ingress_replicationns_command,
            )
        ):
            static_level_cmds = [cmd for cmd in commands if "peer" in cmd]
            parents = [interface_command, vni_command]
            commands = [cmd for cmd in commands if "peer" not in cmd]
            for cmd in commands:
                parents.append(cmd)
            candidate.add(static_level_cmds, parents=parents)

        elif "peer-ip" in commands[0]:
            static_level_cmds = list(commands)
            parents = [
                interface_command,
                vni_command,
                ingress_replications_command,
            ]
            candidate.add(static_level_cmds, parents=parents)

        if vni_command in commands:
            parents = [interface_command]
            commands.remove(vni_command)
            if module.params["assoc_vrf"] is None:
                parents.append(vni_command)
            candidate.add(commands, parents=parents)


def state_absent(module, existing, proposed, candidate):
    if existing["assoc_vrf"]:
        commands = ["no member vni {0} associate-vrf".format(module.params["vni"])]
    else:
        commands = ["no member vni {0}".format(module.params["vni"])]
    parents = ["interface {0}".format(module.params["interface"])]
    candidate.add(commands, parents=parents)


def main():
    argument_spec = dict(
        interface=dict(required=True, type="str"),
        vni=dict(required=True, type="str"),
        assoc_vrf=dict(required=False, type="bool"),
        multicast_group=dict(required=False, type="str"),
        peer_list=dict(required=False, type="list", elements="str"),
        suppress_arp=dict(required=False, type="bool"),
        suppress_arp_disable=dict(required=False, type="bool"),
        ingress_replication=dict(required=False, type="str", choices=["bgp", "static", "default"]),
        state=dict(choices=["present", "absent"], default="present", required=False),
        multisite_ingress_replication=dict(
            required=False,
            type="str",
            choices=["enable", "optimized", "disable"],
        ),
    )

    mutually_exclusive = [
        ("suppress_arp", "suppress_arp_disable"),
        ("assoc_vrf", "multicast_group"),
        ("assoc_vrf", "suppress_arp"),
        ("assoc_vrf", "suppress_arp_disable"),
        ("assoc_vrf", "ingress_replication"),
    ]
    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=mutually_exclusive,
        supports_check_mode=True,
    )

    warnings = list()
    result = {"changed": False, "commands": [], "warnings": warnings}

    if module.params["peer_list"]:
        if (
            module.params["peer_list"][0] != "default"
            and module.params["ingress_replication"] != "static"
        ):
            module.fail_json(
                msg="ingress_replication=static is required " "when using peer_list param",
            )
        else:
            peer_list = module.params["peer_list"]
            if peer_list[0] == "default":
                module.params["peer_list"] = "default"
            else:
                stripped_peer_list = list(map(str.strip, peer_list))
                module.params["peer_list"] = stripped_peer_list

    if (
        module.params["multisite_ingress_replication"] == "enable"
        or module.params["multisite_ingress_replication"] == "optimized"
    ):
        if module.params["ingress_replication"] == "static":
            module.fail_json(
                msg="ingress_replication=static is not allowed "
                "when using multisite_ingress_replication",
            )

    state = module.params["state"]
    args = PARAM_TO_COMMAND_KEYMAP.keys()
    existing, interface_exist = get_existing(module, args)

    if state == "present":
        if not interface_exist:
            module.fail_json(
                msg="The proposed NVE interface does not exist. Use nxos_interface to create it first.",
            )
        elif interface_exist != module.params["interface"]:
            module.fail_json(msg="Only 1 NVE interface is allowed on the switch.")
    elif state == "absent":
        if interface_exist != module.params["interface"]:
            module.exit_json(**result)
        elif existing and existing["vni"] != module.params["vni"]:
            module.fail_json(
                msg="ERROR: VNI delete failed: Could not find vni node for {0}".format(
                    module.params["vni"],
                ),
                existing_vni=existing["vni"],
            )

    proposed_args = dict((k, v) for k, v in module.params.items() if v is not None and k in args)

    proposed = {}
    for key, value in proposed_args.items():
        if key in ["multicast_group", "peer_list", "ingress_replication"]:
            if str(value).lower() == "default":
                value = PARAM_TO_DEFAULT_KEYMAP.get(key, "default")
        if key != "interface" and existing.get(key) != value:
            proposed[key] = value

    candidate = CustomNetworkConfig(indent=3)
    if state == "present":
        state_present(module, existing, proposed, candidate)
    elif existing and state == "absent":
        state_absent(module, existing, proposed, candidate)

    if candidate:
        candidate = candidate.items_text()
        result["changed"] = True
        result["commands"] = candidate
        if not module.check_mode:
            load_config(module, candidate)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Shreyas Srish (@shrsr) <ssrish@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_service_node_type
short_description: Manage Service Node Types
description:
- Manage Service Node Types on Cisco ACI Multi-Site.
author:
- Shreyas Srish (@shrsr)
options:
  name:
    description:
    - The name of the node type.
    type: str
  display_name:
    description:
    - The name of the node type as displayed on the MSO web interface.
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new Service Node Type
  cisco.mso.mso_schema_service_node:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    name: ips
    display_name: ips
    state: present

- name: Remove a Service Node Type
  cisco.mso.mso_schema_service_node:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    name: ips
    state: absent

- name: Query a specific Service Node Type
  cisco.mso.mso_schema_service_node:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    name: ips
    state: query
  register: query_result

- name: Query all Service Node Types
  cisco.mso.mso_schema_service_node:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        name=dict(type="str"),
        display_name=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["name"]],
            ["state", "present", ["name"]],
        ],
    )

    name = module.params.get("name")
    display_name = module.params.get("display_name")
    state = module.params.get("state")

    mso = MSOModule(module)

    mso.existing = {}
    service_node_id = None

    # Get service node id
    query_node_data = mso.query_service_node_types()
    service_nodes = [f.get("name") for f in query_node_data]
    if name in service_nodes:
        for node_data in query_node_data:
            if node_data.get("name") == name:
                service_node_id = node_data.get("id")
                mso.existing = node_data

    if state == "query":
        if name is None:
            mso.existing = query_node_data
        if name is not None and service_node_id is None:
            mso.fail_json(msg="Service Node Type '{service_node_type}' not found".format(service_node_type=name))
        mso.exit_json()

    service_nodes_path = "schemas/service-node-types"
    service_node_path = "schemas/service-node-types/{0}".format(service_node_id)

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing:
            if module.check_mode:
                mso.existing = {}
            else:
                mso.existing = mso.request(service_node_path, method="DELETE")

    elif state == "present":
        if display_name is None:
            display_name = name

        payload = dict(
            name=name,
            displayName=display_name,
        )
        mso.sanitize(payload, collate=True)
        if not module.check_mode:
            if not mso.existing:
                mso.request(service_nodes_path, method="POST", data=payload)
            elif mso.existing.get("displayName") != display_name:
                mso.fail_json(
                    msg="Service Node Type '{0}' already exists with display name '{1}' which is different from provided display name '{2}'.".format(
                        name, mso.existing.get("displayName"), display_name
                    )
                )
        mso.existing = mso.proposed

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2023, Tim Cragg (@timcragg) <tcragg@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_igmp_interface_policy
short_description: Manage IGMP Interface Policies (igmp:IfPol)
description:
- Manage IGMP Interface Policies on Cisco ACI fabrics.
options:
  name:
    description:
    - The name of the IGMP Interface Policy.
    type: str
  tenant:
    description:
    - The tenant to build the IGMP Interface Policy under.
    type: str
  description:
    description:
    - The description of the IGMP Interface Policy.
    type: str
  group_timeout:
    description:
    - The IGMP group timeout in seconds.
    - The APIC defaults to 260 when unset during creation.
    type: int
  query_interval:
    description:
    - The IGMP query interval in seconds.
    - The APIC defaults to 125 when unset during creation.
    type: int
  query_response_interval:
    description:
    - The IGMP query response interval in seconds.
    - The APIC defaults to 10 when unset during creation.
    type: int
  last_member_count:
    description:
    - The last member query count.
    - The APIC defaults to 2 when unset during creation.
    type: int
  last_member_response:
    description:
    - The last member response time in seconds.
    - The APIC defaults to 1 when unset during creation.
    type: int
  startup_query_count:
    description:
    - The Startup Query Count.
    - The APIC defaults to 2 when unset during creation.
    type: int
  startup_query_interval:
    description:
    - The startup query interval in seconds.
    - The APIC defaults to 31 when unset during creation.
    type: int
  querier_timeout:
    description:
    - The querier timeout in seconds.
    - The APIC defaults to 255 when unset during creation.
    type: int
  robustness_variable:
    description:
    - The robustness factor.
    - The APIC defaults to 2 when unset during creation.
    type: int
  igmp_version:
    description:
    - The IGMP version to run.
    - The APIC defaults to v2 when unset during creation.
    type: str
    choices: [ v2, v3 ]
  allow_v3_asm:
    description:
    - Enable the Allow v3 ASM option.
    - The APIC defaults to False when unset during creation.
    - If this parameter is set, fast_leave and report_link_local_groups must also be set.
    type: bool
  fast_leave:
    description:
    - Enable the Fast Leave option.
    - The APIC defaults to False when unset during creation.
    - If this parameter is set, allow_v3_asm and report_link_local_groups must also be set.
    type: bool
  report_link_local_groups:
    description:
    - Enable the Report Link Local Groups option.
    - The APIC defaults to False when unset during creation.
    - If this parameter is set, allow_v3_asm and fast_leave must also be set.
    type: bool
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(igmp:IfPol).
  link: https://developer.cisco.com/docs/apic-mim-ref/

author:
- Tim Cragg (@timcragg)
"""

EXAMPLES = r"""
- name: Add an IGMP Interface Policy
  cisco.aci.aci_igmp_interface_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ans_igmp_intf_pol
    query_interval: 200
    state: present
  delegate_to: localhost

- name: Query an IGMP Interface Policy
  cisco.aci.aci_igmp_interface_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ans_igmp_intf_pol
    state: query
  delegate_to: localhost

- name: Query all IGMP Interface Policies
  cisco.aci.aci_igmp_interface_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost

- name: Remove an IGMP Interface Policy
  cisco.aci.aci_igmp_interface_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ans_igmp_intf_pol
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec, aci_owner_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        name=dict(type="str"),
        tenant=dict(type="str"),
        description=dict(type="str"),
        group_timeout=dict(type="int"),
        query_interval=dict(type="int"),
        query_response_interval=dict(type="int"),
        last_member_count=dict(type="int"),
        last_member_response=dict(type="int"),
        startup_query_count=dict(type="int"),
        startup_query_interval=dict(type="int"),
        querier_timeout=dict(type="int"),
        robustness_variable=dict(type="int"),
        igmp_version=dict(type="str", choices=["v2", "v3"]),
        allow_v3_asm=dict(type="bool"),
        fast_leave=dict(type="bool"),
        report_link_local_groups=dict(type="bool"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["name", "tenant"]],
            ["state", "absent", ["name", "tenant"]],
        ],
        required_together=[
            ["allow_v3_asm", "fast_leave", "report_link_local_groups"],
        ],
    )

    aci = ACIModule(module)

    name = module.params.get("name")
    tenant = module.params.get("tenant")
    description = module.params.get("description")
    group_timeout = module.params.get("group_timeout")
    query_interval = module.params.get("query_interval")
    query_response_interval = module.params.get("query_response_interval")
    last_member_count = module.params.get("last_member_count")
    last_member_response = module.params.get("last_member_response")
    startup_query_count = module.params.get("startup_query_count")
    startup_query_interval = module.params.get("startup_query_interval")
    querier_timeout = module.params.get("querier_timeout")
    robustness_variable = module.params.get("robustness_variable")
    igmp_version = module.params.get("igmp_version")
    allow_v3_asm = module.params.get("allow_v3_asm")
    fast_leave = module.params.get("fast_leave")
    report_link_local_groups = module.params.get("report_link_local_groups")
    state = module.params.get("state")

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="igmpIfPol",
            aci_rn="igmpIfPol-{0}".format(name),
            module_object=name,
            target_filter={"name": name},
        ),
    )
    aci.get_existing()

    if state == "present":
        if allow_v3_asm is not None:
            if_ctrl_list = []
            if allow_v3_asm:
                if_ctrl_list.append("allow-v3-asm")
            if fast_leave:
                if_ctrl_list.append("fast-leave")
            if report_link_local_groups:
                if_ctrl_list.append("rep-ll")
            if_ctrl = ",".join(if_ctrl_list)
        else:
            if_ctrl = None

        aci.payload(
            aci_class="igmpIfPol",
            class_config=dict(
                name=name,
                descr=description,
                grpTimeout=group_timeout,
                ifCtrl=if_ctrl,
                lastMbrCnt=last_member_count,
                lastMbrRespTime=last_member_response,
                querierTimeout=querier_timeout,
                queryIntvl=query_interval,
                robustFac=robustness_variable,
                rspIntvl=query_response_interval,
                startQueryCnt=startup_query_count,
                startQueryIntvl=startup_query_interval,
                ver=igmp_version,
            ),
        )

        aci.get_diff(aci_class="igmpIfPol")
        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: cloudformation
version_added: 1.0.0
short_description: Create or delete an AWS CloudFormation stack
description:
  - Launches or updates an AWS CloudFormation stack and waits for it complete.
options:
  stack_name:
    description:
      - Name of the CloudFormation stack.
    required: true
    type: str
  disable_rollback:
    description:
      - If a stacks fails to form, rollback will remove the stack.
    default: false
    type: bool
  on_create_failure:
    description:
      - Action to take upon failure of stack creation. Incompatible with the O(disable_rollback) option.
    choices:
      - DO_NOTHING
      - ROLLBACK
      - DELETE
    type: str
  create_timeout:
    description:
      - The amount of time (in minutes) that can pass before the stack status becomes V(CREATE_FAILED).
    type: int
  template_parameters:
    description:
      - A list of hashes of all the template variables for the stack. The value can be a string or a dict.
      - Dict can be used to set additional template parameter attributes like UsePreviousValue (see example).
    default: {}
    type: dict
  state:
    description:
      - If O(state=present), stack will be created.
      - If O(state=present) and if stack exists and template has changed, it will be updated.
      - If O(state=absent), stack will be removed.
    default: present
    choices: [ present, absent ]
    type: str
  template:
    description:
      - The local path of the CloudFormation template.
      - This must be the full path to the file, relative to the working directory. If using roles this may look
        like V(roles/cloudformation/files/cloudformation-example.json).
      - If O(state=present) and the stack does not exist yet, either O(template), O(template_body) or O(template_url)
        must be specified (but only one of them).
      - If O(state=present), the stack does exist, and neither O(template),
        O(template_body) nor O(template_url) are specified, the previous template will be reused.
      - The O(template) parameter has been deprecated and will be remove in a release after
        2026-05-01. It is recommended to use O(template_body) with the P(ansible.builtin.template#lookup) lookup plugin.
    type: path
  notification_arns:
    description:
      - A comma separated list of Simple Notification Service (SNS) topic ARNs to publish stack related events.
    type: str
  stack_policy:
    description:
      - The path of the file containing the CloudFormation stack policy. A policy cannot be removed once placed, but it can be modified.
        for instance, allow all updates U(https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/protect-stack-resources.html#d0e9051)
    type: str
  stack_policy_body:
    description:
      - The CloudFormation stack policy in JSON. A policy cannot be removed once placed, but it can be modified.
        for instance, allow all updates U(https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/protect-stack-resources.html#d0e9051)
    type: json
    version_added: 1.5.0
  stack_policy_on_update_body:
    description:
      - The body of the cloudformation stack policy only applied during this update.
    type: json
    version_added: 1.5.0
  tags:
    description:
      - Dictionary of tags to associate with stack and its resources during stack creation.
      - Can be updated later, updating tags removes previous entries.
    type: dict
  template_url:
    description:
      - Location of file containing the template body. The URL must point to a template (max size 307,200 bytes) located in an
        S3 bucket in the same region as the stack.
      - If O(state=present) and the stack does not exist yet, either O(template), O(template_body) or O(template_url)
        must be specified (but only one of them).
      - If O(state=present), the stack does exist, and neither O(template), O(template_body) nor O(template_url) are specified,
        the previous template will be reused.
    type: str
  create_changeset:
    description:
      - If stack already exists create a changeset instead of directly applying changes.  See the AWS Change Sets docs
        U(https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/using-cfn-updating-stacks-changesets.html).
      - B(WARNING:) if the stack does not exist, it will be created without changeset. If O(state=absent), the stack will be
        deleted immediately with no changeset.
    type: bool
    default: false
  changeset_name:
    description:
      - Name given to the changeset when creating a changeset.
      - Only used when O(create_changeset=true).
      - By default a name prefixed with Ansible-STACKNAME is generated based on input parameters.
        See the AWS Change Sets docs for more information
        U(https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/using-cfn-updating-stacks-changesets.html)
    type: str
  role_arn:
    description:
      - The role that AWS CloudFormation assumes to create the stack. See the AWS CloudFormation Service Role
        docs U(https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/using-iam-servicerole.html)
    type: str
  termination_protection:
    description:
      - Enable or disable termination protection on the stack.
    type: bool
  template_body:
    description:
      - Template body. Use this to pass in the actual body of the CloudFormation template.
      - If O(state=present) and the stack does not exist yet, either O(template), O(template_body) or O(template_url)
        must be specified (but only one of them).
      - If O(state=present), the stack does exist, and neither O(template), O(template_body) nor O(template_url)
        are specified, the previous template will be reused.
    type: str
  events_limit:
    description:
      - Maximum number of CloudFormation events to fetch from a stack when creating or updating it.
    default: 200
    type: int
  backoff_delay:
    description:
      - Number of seconds to wait for the next retry.
    default: 3
    type: int
    required: False
  backoff_max_delay:
    description:
      - Maximum amount of time to wait between retries.
    default: 30
    type: int
    required: False
  backoff_retries:
    description:
      - Number of times to retry operation.
      - AWS API throttling mechanism fails CloudFormation module so we have to retry a couple of times.
    default: 10
    type: int
    required: False
  capabilities:
    description:
      - Specify capabilities that stack template contains.
      - Valid values are V(CAPABILITY_IAM), V(CAPABILITY_NAMED_IAM) and V(CAPABILITY_AUTO_EXPAND).
    type: list
    elements: str
    default: [ CAPABILITY_IAM, CAPABILITY_NAMED_IAM ]

author:
  - "James S. Martin (@jsmartin)"
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: create a cloudformation stack
  amazon.aws.cloudformation:
    stack_name: "ansible-cloudformation"
    state: "present"
    region: "us-east-1"
    disable_rollback: true
    # The template parameter has been deprecated, use template_body with lookup instead.
    # template: "files/cloudformation-example.json"
    template_body: "{{ lookup('file', 'cloudformation-example.json') }}"
    template_parameters:
      KeyName: "jmartin"
      DiskType: "ephemeral"
      InstanceType: "m1.small"
      ClusterSize: 3
    tags:
      Stack: "ansible-cloudformation"

# Basic role example
- name: create a stack, specify role that cloudformation assumes
  amazon.aws.cloudformation:
    stack_name: "ansible-cloudformation"
    state: "present"
    region: "us-east-1"
    disable_rollback: true
    # The template parameter has been deprecated, use template_body with lookup instead.
    # template: "roles/cloudformation/files/cloudformation-example.json"
    template_body: "{{ lookup('file', 'cloudformation-example.json') }}"
    role_arn: 'arn:aws:iam::123456789012:role/cloudformation-iam-role'

- name: delete a stack
  amazon.aws.cloudformation:
    stack_name: "ansible-cloudformation-old"
    state: "absent"

# Create a stack, pass in template from a URL, disable rollback if stack creation fails,
# pass in some parameters to the template, provide tags for resources created
- name: create a stack, pass in the template via an URL
  amazon.aws.cloudformation:
    stack_name: "ansible-cloudformation"
    state: present
    region: us-east-1
    disable_rollback: true
    template_url: https://s3.amazonaws.com/my-bucket/cloudformation.template
    template_parameters:
      KeyName: jmartin
      DiskType: ephemeral
      InstanceType: m1.small
      ClusterSize: 3
    tags:
      Stack: ansible-cloudformation

# Create a stack, passing in template body using lookup of Jinja2 template, disable rollback if stack creation fails,
# pass in some parameters to the template, provide tags for resources created
- name: create a stack, pass in the template body via lookup template
  amazon.aws.cloudformation:
    stack_name: "ansible-cloudformation"
    state: present
    region: us-east-1
    disable_rollback: true
    template_body: "{{ lookup('template', 'cloudformation.j2') }}"
    template_parameters:
      KeyName: jmartin
      DiskType: ephemeral
      InstanceType: m1.small
      ClusterSize: 3
    tags:
      Stack: ansible-cloudformation

# Pass a template parameter which uses CloudFormation's UsePreviousValue attribute
# When use_previous_value is set to True, the given value will be ignored and
# CloudFormation will use the value from a previously submitted template.
# If use_previous_value is set to False (default) the given value is used.
- amazon.aws.cloudformation:
    stack_name: "ansible-cloudformation"
    state: "present"
    region: "us-east-1"
    template: "files/cloudformation-example.json"
    template_parameters:
      DBSnapshotIdentifier:
        use_previous_value: true
        value: arn:aws:rds:es-east-1:123456789012:snapshot:rds:my-db-snapshot
      DBName:
        use_previous_value: true
    tags:
      Stack: "ansible-cloudformation"

# Enable termination protection on a stack.
# If the stack already exists, this will update its termination protection
- name: enable termination protection during stack creation
  amazon.aws.cloudformation:
    stack_name: my_stack
    state: present
    template_url: https://s3.amazonaws.com/my-bucket/cloudformation.template
    termination_protection: true

# Configure TimeoutInMinutes before the stack status becomes CREATE_FAILED
# In this case, if disable_rollback is not set or is set to false, the stack will be rolled back.
- name: enable termination protection during stack creation
  amazon.aws.cloudformation:
    stack_name: my_stack
    state: present
    template_url: https://s3.amazonaws.com/my-bucket/cloudformation.template
    create_timeout: 5

# Configure rollback behaviour on the unsuccessful creation of a stack allowing
# CloudFormation to clean up, or do nothing in the event of an unsuccessful
# deployment
# In this case, if on_create_failure is set to "DELETE", it will clean up the stack if
# it fails to create
- name: create stack which will delete on creation failure
  amazon.aws.cloudformation:
    stack_name: my_stack
    state: present
    template_url: https://s3.amazonaws.com/my-bucket/cloudformation.template
    on_create_failure: DELETE
"""

RETURN = r"""
events:
  description: Most recent events in CloudFormation's event log. This may be from a previous run in some cases.
  returned: always
  type: list
  sample: [
        "StackEvent AWS::CloudFormation::Stack stackname UPDATE_COMPLETE",
        "StackEvent AWS::CloudFormation::Stack stackname UPDATE_COMPLETE_CLEANUP_IN_PROGRESS"
    ]
log:
  description: Debugging logs. Useful when modifying or finding an error.
  returned: always
  type: list
  elements: str
  sample: ["updating stack"]
change_set_id:
  description: The ID of the stack change set when created.
  returned:  when O(state=present) and O(create_changeset=true).
  type: str
  sample: "arn:aws:cloudformation:us-east-1:123456789012:changeSet/Ansible-StackName-f4496805bd1b2be824d1e315c6884247ede41eb0"
stack_resources:
  description: AWS stack resources and their status. List of dictionaries, one dict per resource.
  returned: when O(state=present)
  type: list
  elements: dict
  sample: [
          {
              "last_updated_time": "2016-10-11T19:40:14.979000+00:00",
              "logical_resource_id": "CFTestSg",
              "physical_resource_id": "cloudformation2-CFTestSg-16UQ4CYQ57O9F",
              "resource_type": "AWS::EC2::SecurityGroup",
              "status": "UPDATE_COMPLETE",
              "status_reason": null
          }
      ]
stack_outputs:
  description: A key:value dictionary of all the stack outputs currently defined. If there are no stack outputs, it is an empty dictionary.
  returned: When O(state=present)
  type: dict
  sample: {"MySg": "AnsibleModuleTestYAML-CFTestSg-C8UVS567B6NS"}
"""

import json
import time
import traceback
import uuid
from hashlib import sha1

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.text.converters import to_bytes
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.amazon.aws.plugins.module_utils.botocore import boto_exception
from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_message
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list


@AWSRetry.jittered_backoff()
def _search_events(cfn, stack_name, events_limit, token_filter):
    pg = cfn.get_paginator("describe_stack_events").paginate(
        StackName=stack_name,
        PaginationConfig={"MaxItems": events_limit},
    )
    if token_filter is None:
        return list(pg.search("StackEvents[*]"))

    return list(pg.search(f"StackEvents[?ClientRequestToken == '{token_filter}']"))


def get_stack_events(cfn, stack_name, events_limit, token_filter=None):
    """This event data was never correct, it worked as a side effect. So the v2.3 format is different."""
    ret = {"events": [], "log": []}

    try:
        events = _search_events(cfn, stack_name, events_limit, token_filter)
    except is_boto3_error_message("does not exist"):
        ret["log"].append("Stack does not exist.")
        return ret
    except (
        botocore.exceptions.ValidationError,
        botocore.exceptions.ClientError,
    ) as err:  # pylint: disable=duplicate-except
        error_msg = boto_exception(err)
        ret["log"].append("Unknown error: " + str(error_msg))
        return ret

    for e in events:
        eventline = f"StackEvent {e['ResourceType']} {e['LogicalResourceId']} {e['ResourceStatus']}"
        ret["events"].append(eventline)

        if e["ResourceStatus"].endswith("FAILED"):
            failure = f"{e['ResourceType']} {e['LogicalResourceId']} {e['ResourceStatus']}: {e['ResourceStatusReason']}"
            ret["log"].append(failure)

    return ret


def create_stack(module, stack_params, cfn, events_limit):
    if "TemplateBody" not in stack_params and "TemplateURL" not in stack_params:
        module.fail_json(
            msg="Either 'template', 'template_body' or 'template_url' is required when the stack does not exist."
        )

    # 'TimeoutInMinutes', 'EnableTerminationProtection' and
    # 'OnFailure' only apply on creation, not update.
    if module.params.get("on_create_failure") is not None:
        stack_params["OnFailure"] = module.params["on_create_failure"]
    else:
        stack_params["DisableRollback"] = module.params["disable_rollback"]

    if module.params.get("create_timeout") is not None:
        stack_params["TimeoutInMinutes"] = module.params["create_timeout"]
    if module.params.get("termination_protection") is not None:
        stack_params["EnableTerminationProtection"] = bool(module.params.get("termination_protection"))

    try:
        response = cfn.create_stack(aws_retry=True, **stack_params)
        # Use stack ID to follow stack state in case of on_create_failure = DELETE
        result = stack_operation(
            module, cfn, response["StackId"], "CREATE", events_limit, stack_params.get("ClientRequestToken", None)
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as err:
        module.fail_json_aws(err, msg=f"Failed to create stack {stack_params.get('StackName')}")
    if not result:
        module.fail_json(msg="empty result")
    return result


def list_changesets(cfn, stack_name):
    res = cfn.list_change_sets(aws_retry=True, StackName=stack_name)
    return [cs["ChangeSetName"] for cs in res["Summaries"]]


def create_changeset(module, stack_params, cfn, events_limit):
    if "TemplateBody" not in stack_params and "TemplateURL" not in stack_params:
        module.fail_json(msg="Either 'template' or 'template_url' is required.")
    if module.params["changeset_name"] is not None:
        stack_params["ChangeSetName"] = module.params["changeset_name"]

    # changesets don't accept ClientRequestToken parameters
    stack_params.pop("ClientRequestToken", None)

    try:
        changeset_name = build_changeset_name(stack_params)
        stack_params["ChangeSetName"] = changeset_name

        # Determine if this changeset already exists
        pending_changesets = list_changesets(cfn, stack_params["StackName"])
        if changeset_name in pending_changesets:
            warning = f"WARNING: {len(pending_changesets)} pending changeset(s) exist(s) for this stack!"
            result = dict(changed=False, output=f"ChangeSet {changeset_name} already exists.", warnings=[warning])
        else:
            cs = cfn.create_change_set(aws_retry=True, **stack_params)
            # Make sure we don't enter an infinite loop
            time_end = time.time() + 600
            while time.time() < time_end:
                try:
                    newcs = cfn.describe_change_set(aws_retry=True, ChangeSetName=cs["Id"])
                except botocore.exceptions.BotoCoreError as err:
                    module.fail_json_aws(err)
                if newcs["Status"] == "CREATE_PENDING" or newcs["Status"] == "CREATE_IN_PROGRESS":
                    time.sleep(1)
                elif newcs["Status"] == "FAILED" and (
                    "The submitted information didn't contain changes" in newcs["StatusReason"]
                    or "No updates are to be performed" in newcs["StatusReason"]
                ):
                    cfn.delete_change_set(aws_retry=True, ChangeSetName=cs["Id"])
                    result = dict(
                        changed=False,
                        output="The created Change Set did not contain any changes to this stack and was deleted.",
                    )
                    # a failed change set does not trigger any stack events so we just want to
                    # skip any further processing of result and just return it directly
                    return result
                else:
                    break
                # Lets not hog the cpu/spam the AWS API
                time.sleep(1)
            result = stack_operation(module, cfn, stack_params["StackName"], "CREATE_CHANGESET", events_limit)
            result["change_set_id"] = cs["Id"]
            result["warnings"] = [
                f"Created changeset named {changeset_name} for stack {stack_params['StackName']}",
                f"You can execute it using: aws cloudformation execute-change-set --change-set-name {cs['Id']}",
                "NOTE that dependencies on this stack might fail due to pending changes!",
            ]
    except is_boto3_error_message("No updates are to be performed."):
        result = dict(changed=False, output="Stack is already up-to-date.")
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as err:
        module.fail_json_aws(err, msg="Failed to create change set")

    if not result:
        module.fail_json(msg="empty result")
    return result


def update_stack(module, stack_params, cfn, events_limit):
    if "TemplateBody" not in stack_params and "TemplateURL" not in stack_params:
        stack_params["UsePreviousTemplate"] = True

    if module.params["stack_policy_on_update_body"] is not None:
        stack_params["StackPolicyDuringUpdateBody"] = module.params["stack_policy_on_update_body"]

    stack_params["DisableRollback"] = module.params["disable_rollback"]

    # if the state is present and the stack already exists, we try to update it.
    # AWS will tell us if the stack template and parameters are the same and
    # don't need to be updated.
    try:
        cfn.update_stack(aws_retry=True, **stack_params)
        result = stack_operation(
            module, cfn, stack_params["StackName"], "UPDATE", events_limit, stack_params.get("ClientRequestToken", None)
        )
    except is_boto3_error_message("No updates are to be performed."):
        result = dict(changed=False, output="Stack is already up-to-date.")
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as err:
        module.fail_json_aws(err, msg=f"Failed to update stack {stack_params.get('StackName')}")
    if not result:
        module.fail_json(msg="empty result")
    return result


def update_termination_protection(module, cfn, stack_name, desired_termination_protection_state):
    """updates termination protection of a stack"""
    changed = False
    stack = get_stack_facts(module, cfn, stack_name)
    if stack:
        if stack["EnableTerminationProtection"] is not desired_termination_protection_state:
            try:
                cfn.update_termination_protection(
                    aws_retry=True,
                    EnableTerminationProtection=desired_termination_protection_state,
                    StackName=stack_name,
                )
                changed = True
            except botocore.exceptions.ClientError as e:
                module.fail_json_aws(e)
    return changed


def stack_operation(module, cfn, stack_name, operation, events_limit, op_token=None):
    """gets the status of a stack while it is created/updated/deleted"""
    existed = []
    while True:
        try:
            stack = get_stack_facts(module, cfn, stack_name, raise_errors=True)
            existed.append("yes")
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError):
            # If the stack previously existed, and now can't be found then it's
            # been deleted successfully.
            if "yes" in existed or operation == "DELETE":  # stacks may delete fast, look in a few ways.
                ret = get_stack_events(cfn, stack_name, events_limit, op_token)
                ret.update({"changed": True, "output": "Stack Deleted"})
                return ret
            else:
                return {
                    "changed": True,
                    "failed": True,
                    "output": "Stack Not Found",
                    "exception": traceback.format_exc(),
                }
        ret = get_stack_events(cfn, stack_name, events_limit, op_token)
        if not stack:
            if "yes" in existed or operation == "DELETE":  # stacks may delete fast, look in a few ways.
                ret = get_stack_events(cfn, stack_name, events_limit, op_token)
                ret.update({"changed": True, "output": "Stack Deleted"})
                return ret
            else:
                ret.update({"changed": False, "failed": True, "output": "Stack not found."})
                return ret
        # it covers ROLLBACK_COMPLETE and UPDATE_ROLLBACK_COMPLETE
        # Possible states: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/using-cfn-describing-stacks.html#w1ab2c15c17c21c13
        elif stack["StackStatus"].endswith("ROLLBACK_COMPLETE") and operation != "CREATE_CHANGESET":
            ret.update({"changed": True, "failed": True, "output": f"Problem with {operation}. Rollback complete"})
            return ret
        elif stack["StackStatus"] == "DELETE_COMPLETE" and operation == "CREATE":
            ret.update({"changed": True, "failed": True, "output": "Stack create failed. Delete complete."})
            return ret
        # note the ordering of ROLLBACK_COMPLETE, DELETE_COMPLETE, and COMPLETE, because otherwise COMPLETE will match all cases.
        elif stack["StackStatus"].endswith("_COMPLETE"):
            ret.update({"changed": True, "output": f"Stack {operation} complete"})
            return ret
        elif stack["StackStatus"].endswith("_ROLLBACK_FAILED"):
            ret.update({"changed": True, "failed": True, "output": f"Stack {operation} rollback failed"})
            return ret
        # note the ordering of ROLLBACK_FAILED and FAILED, because otherwise FAILED will match both cases.
        elif stack["StackStatus"].endswith("_FAILED"):
            ret.update({"changed": True, "failed": True, "output": f"Stack {operation} failed"})
            return ret
        else:
            # this can loop forever :/
            time.sleep(5)
    return {"failed": True, "output": "Failed for unknown reasons."}


def build_changeset_name(stack_params):
    if "ChangeSetName" in stack_params:
        return stack_params["ChangeSetName"]

    json_params = json.dumps(stack_params, sort_keys=True)

    changeset_sha = sha1(to_bytes(json_params, errors="surrogate_or_strict")).hexdigest()
    return f"Ansible-{stack_params['StackName']}-{changeset_sha}"


def check_mode_changeset(module, stack_params, cfn):
    """Create a change set, describe it and delete it before returning check mode outputs."""
    stack_params["ChangeSetName"] = build_changeset_name(stack_params)
    # changesets don't accept ClientRequestToken parameters
    stack_params.pop("ClientRequestToken", None)

    try:
        for _i in range(60):  # total time 5 min
            # check stack is ready to have a change set created
            stack = get_stack_facts(module, cfn, stack_params["StackName"], raise_errors=True)
            if stack["StackStatus"].endswith("_COMPLETE"):
                break
            time.sleep(5)
        change_set = cfn.create_change_set(aws_retry=True, **stack_params)
        for _i in range(60):  # total time 5 min
            description = cfn.describe_change_set(aws_retry=True, ChangeSetName=change_set["Id"])
            if description["Status"] in ("CREATE_COMPLETE", "FAILED"):
                break
            time.sleep(5)
        else:
            # if the changeset doesn't finish in 5 mins, this `else` will trigger and fail
            module.fail_json(msg=f"Failed to create change set {stack_params['ChangeSetName']}")

        cfn.delete_change_set(aws_retry=True, ChangeSetName=change_set["Id"])

        reason = description.get("StatusReason")

        if description["Status"] == "FAILED" and (
            "didn't contain changes" in reason or "No updates are to be performed" in reason
        ):
            return {"changed": False, "msg": reason, "meta": reason}
        return {"changed": True, "msg": reason, "meta": description["Changes"]}

    except (botocore.exceptions.ValidationError, botocore.exceptions.ClientError) as err:
        module.fail_json_aws(err)


def get_stack_facts(module, cfn, stack_name, raise_errors=False):
    try:
        stack_response = cfn.describe_stacks(aws_retry=True, StackName=stack_name)
        stack_info = stack_response["Stacks"][0]
    except is_boto3_error_message("does not exist"):
        return None
    except (
        botocore.exceptions.ValidationError,
        botocore.exceptions.ClientError,
    ) as err:  # pylint: disable=duplicate-except
        if raise_errors:
            raise err
        module.fail_json_aws(err, msg="Failed to describe stack")

    if stack_response and stack_response.get("Stacks", None):
        stacks = stack_response["Stacks"]
        if len(stacks):
            stack_info = stacks[0]

    return stack_info


def main():
    argument_spec = dict(
        stack_name=dict(required=True),
        template_parameters=dict(required=False, type="dict", default={}),
        state=dict(default="present", choices=["present", "absent"]),
        template=dict(
            default=None,
            required=False,
            type="path",
            removed_at_date="2026-05-01",
            removed_from_collection="amazon.aws",
        ),
        notification_arns=dict(default=None, required=False),
        stack_policy=dict(default=None, required=False),
        stack_policy_body=dict(default=None, required=False, type="json"),
        stack_policy_on_update_body=dict(default=None, required=False, type="json"),
        disable_rollback=dict(default=False, type="bool"),
        on_create_failure=dict(default=None, required=False, choices=["DO_NOTHING", "ROLLBACK", "DELETE"]),
        create_timeout=dict(default=None, type="int"),
        template_url=dict(default=None, required=False),
        template_body=dict(default=None, required=False),
        create_changeset=dict(default=False, type="bool"),
        changeset_name=dict(default=None, required=False),
        role_arn=dict(default=None, required=False),
        tags=dict(default=None, type="dict"),
        termination_protection=dict(default=None, type="bool"),
        events_limit=dict(default=200, type="int"),
        backoff_retries=dict(type="int", default=10, required=False),
        backoff_delay=dict(type="int", default=3, required=False),
        backoff_max_delay=dict(type="int", default=30, required=False),
        capabilities=dict(type="list", elements="str", default=["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"]),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        mutually_exclusive=[["template_url", "template", "template_body"], ["disable_rollback", "on_create_failure"]],
        supports_check_mode=True,
    )

    invalid_capabilities = []
    user_capabilities = module.params.get("capabilities")
    for user_cap in user_capabilities:
        if user_cap not in ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"]:
            invalid_capabilities.append(user_cap)

    if invalid_capabilities:
        module.fail_json(
            msg=f"Specified capabilities are invalid : {invalid_capabilities!r}, please check documentation for valid capabilities"
        )

    # collect the parameters that are passed to boto3. Keeps us from having so many scalars floating around.
    stack_params = {
        "Capabilities": user_capabilities,
        "ClientRequestToken": to_native(uuid.uuid4()),
    }
    state = module.params["state"]
    stack_params["StackName"] = module.params["stack_name"]

    if module.params["template"] is not None:
        with open(module.params["template"], "r") as template_fh:
            stack_params["TemplateBody"] = template_fh.read()
    elif module.params["template_body"] is not None:
        stack_params["TemplateBody"] = module.params["template_body"]
    elif module.params["template_url"] is not None:
        stack_params["TemplateURL"] = module.params["template_url"]

    if module.params.get("notification_arns"):
        stack_params["NotificationARNs"] = module.params["notification_arns"].split(",")
    else:
        stack_params["NotificationARNs"] = []

    # can't check the policy when verifying.
    if (
        module.params["stack_policy_body"] is not None
        and not module.check_mode
        and not module.params["create_changeset"]
    ):
        stack_params["StackPolicyBody"] = module.params["stack_policy_body"]
    elif module.params["stack_policy"] is not None and not module.check_mode and not module.params["create_changeset"]:
        with open(module.params["stack_policy"], "r") as stack_policy_fh:
            stack_params["StackPolicyBody"] = stack_policy_fh.read()

    template_parameters = module.params["template_parameters"]

    stack_params["Parameters"] = []
    for k, v in template_parameters.items():
        if isinstance(v, dict):
            # set parameter based on a dict to allow additional CFN Parameter Attributes
            param = dict(ParameterKey=k)

            if "value" in v:
                param["ParameterValue"] = str(v["value"])

            if "use_previous_value" in v and bool(v["use_previous_value"]):
                param["UsePreviousValue"] = True
                param.pop("ParameterValue", None)

            stack_params["Parameters"].append(param)
        else:
            # allow default k/v configuration to set a template parameter
            stack_params["Parameters"].append({"ParameterKey": k, "ParameterValue": str(v)})

    if isinstance(module.params.get("tags"), dict):
        stack_params["Tags"] = ansible_dict_to_boto3_tag_list(module.params["tags"])

    if module.params.get("role_arn"):
        stack_params["RoleARN"] = module.params["role_arn"]

    result = {}

    # Wrap the cloudformation client methods that this module uses with
    # automatic backoff / retry for throttling error codes
    retry_decorator = AWSRetry.jittered_backoff(
        retries=module.params.get("backoff_retries"),
        delay=module.params.get("backoff_delay"),
        max_delay=module.params.get("backoff_max_delay"),
    )
    cfn = module.client("cloudformation", retry_decorator=retry_decorator)

    stack_info = get_stack_facts(module, cfn, stack_params["StackName"])

    if module.check_mode:
        if state == "absent" and stack_info:
            module.exit_json(changed=True, msg="Stack would be deleted", meta=[])
        elif state == "absent" and not stack_info:
            module.exit_json(changed=False, msg="Stack doesn't exist", meta=[])
        elif state == "present" and not stack_info:
            module.exit_json(changed=True, msg="New stack would be created", meta=[])
        else:
            module.exit_json(**check_mode_changeset(module, stack_params, cfn))

    if state == "present":
        if not stack_info:
            result = create_stack(module, stack_params, cfn, module.params.get("events_limit"))
        else:
            changeset_updated = False
            if module.params.get("create_changeset"):
                result = create_changeset(module, stack_params, cfn, module.params.get("events_limit"))
                changeset_updated = True
            if module.params.get("termination_protection") is not None:
                result["changed"] = update_termination_protection(
                    module, cfn, stack_params["StackName"], bool(module.params.get("termination_protection"))
                )
            if not changeset_updated:
                result = update_stack(module, stack_params, cfn, module.params.get("events_limit"))

        # format the stack output

        stack = get_stack_facts(module, cfn, stack_params["StackName"])
        if stack is not None:
            if result.get("stack_outputs") is None:
                # always define stack_outputs, but it may be empty
                result["stack_outputs"] = {}
            for output in stack.get("Outputs", []):
                result["stack_outputs"][output["OutputKey"]] = output["OutputValue"]
            stack_resources = []
            reslist = cfn.list_stack_resources(aws_retry=True, StackName=stack_params["StackName"])
            for res in reslist.get("StackResourceSummaries", []):
                stack_resources.append(
                    {
                        "logical_resource_id": res["LogicalResourceId"],
                        "physical_resource_id": res.get("PhysicalResourceId", ""),
                        "resource_type": res["ResourceType"],
                        "last_updated_time": res["LastUpdatedTimestamp"],
                        "status": res["ResourceStatus"],
                        "status_reason": res.get("ResourceStatusReason"),  # can be blank, apparently
                    }
                )
            result["stack_resources"] = stack_resources

    elif state == "absent":
        # absent state is different because of the way delete_stack works.
        # problem is it it doesn't give an error if stack isn't found
        # so must describe the stack first

        try:
            stack = get_stack_facts(module, cfn, stack_params["StackName"])
            if not stack:
                result = {"changed": False, "output": "Stack not found."}
            else:
                if stack_params.get("RoleARN") is None:
                    cfn.delete_stack(aws_retry=True, StackName=stack_params["StackName"])
                else:
                    cfn.delete_stack(
                        aws_retry=True, StackName=stack_params["StackName"], RoleARN=stack_params["RoleARN"]
                    )
                result = stack_operation(
                    module,
                    cfn,
                    stack_params["StackName"],
                    "DELETE",
                    module.params.get("events_limit"),
                    stack_params.get("ClientRequestToken", None),
                )
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as err:
            module.fail_json_aws(err)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import functools
import inspect

try:
    import botocore
except ImportError:
    pass  # Modules are responsible for handling this.

from .exceptions import AnsibleAWSError


class AWSErrorHandler:
    """_CUSTOM_EXCEPTION can be overridden by subclasses to customize the exception raised"""

    _CUSTOM_EXCEPTION = AnsibleAWSError

    @classmethod
    def _is_missing(cls):
        """Should be overridden with a class method that returns the value from is_boto3_error_code (or similar)"""
        return type("NeverEverRaisedException", (Exception,), {})

    @classmethod
    def common_error_handler(cls, description):
        """A simple error handler that catches the standard Boto3 exceptions and raises
        an AnsibleAWSError exception.

        param: description: a description of the action being taken.
                            Exception raised will include a message of
                            f"Timeout trying to {description}" or
                            f"Failed to {description}"
                            Supports format string placeholders matching function parameters,
                            e.g., "get role {name}" will interpolate the 'name' parameter.
        """

        def wrapper(func):
            # Unwrap nested decorators to find the original function signature
            original_func = func
            while hasattr(original_func, "__wrapped__"):
                original_func = original_func.__wrapped__
            sig = inspect.signature(original_func)

            @functools.wraps(func)
            def handler(*args, **kwargs):
                try:
                    bound = sig.bind(*args, **kwargs)
                    bound.apply_defaults()
                    desc = description.format(**bound.arguments)
                except (KeyError, ValueError, AttributeError) as format_error:
                    # Formatting failed - append a note to help debug the issue
                    desc = f"{description} (error formatting message: {format_error})"

                try:
                    return func(*args, **kwargs)
                except botocore.exceptions.WaiterError as e:
                    raise cls._CUSTOM_EXCEPTION(message=f"Timeout trying to {desc}", exception=e) from e
                except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
                    raise cls._CUSTOM_EXCEPTION(message=f"Failed to {desc}", exception=e) from e

            return handler

        return wrapper

    @classmethod
    def list_error_handler(cls, description, default_value=None):
        """A simple error handler that catches the standard Boto3 exceptions and raises
        an AnsibleAWSError exception.
        Error codes representing a non-existent entity will result in None being returned
        Generally used for Get/List calls where the exception just means the resource isn't there

        param: description: a description of the action being taken.
                            Exception raised will include a message of
                            f"Timeout trying to {description}" or
                            f"Failed to {description}"
                            Supports format string placeholders matching function parameters,
                            e.g., "get role {name}" will interpolate the 'name' parameter.
        param: default_value: the value to return if no matching
                            resources are returned.  Defaults to None
        """

        def wrapper(func):
            @cls.common_error_handler(description)
            @functools.wraps(func)
            def handler(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except cls._is_missing():
                    return default_value

            return handler

        return wrapper

    @classmethod
    def deletion_error_handler(cls, description):
        """A simple error handler that catches the standard Boto3 exceptions and raises
        an AnsibleAWSError exception.
        Error codes representing a non-existent entity will result in None being returned
        Generally used in deletion calls where NoSuchEntity means it's already gone

        param: description: a description of the action being taken.
                            Exception raised will include a message of
                            f"Timeout trying to {description}" or
                            f"Failed to {description}"
                            Supports format string placeholders matching function parameters,
                            e.g., "delete role {name}" will interpolate the 'name' parameter.
        """

        def wrapper(func):
            @cls.common_error_handler(description)
            @functools.wraps(func)
            def handler(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except cls._is_missing():
                    return False

            return handler

        return wrapper



# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Standard files documentation fragment
    DOCUMENTATION = r"""
requirements:
  - cryptography >= 3.3
attributes:
  diff_mode:
    support: none
  idempotent:
    support: full
options:
  src_path:
    description:
      - Name of the file containing the OpenSSL private key to convert.
      - Exactly one of O(src_path) or O(src_content) must be specified.
    type: path
  src_content:
    description:
      - The content of the file containing the OpenSSL private key to convert.
      - Exactly one of O(src_path) or O(src_content) must be specified.
    type: str
  src_passphrase:
    description:
      - The passphrase for the private key to load.
    type: str
  dest_passphrase:
    description:
      - The passphrase for the private key to store.
    type: str
  format:
    description:
      - Determines which format the destination private key should be written in.
      - Please note that not every key can be exported in any format, and that not every format supports encryption.
    type: str
    choices: [pkcs1, pkcs8, raw]
    required: true
seealso:
  - module: community.crypto.openssl_privatekey
  - module: community.crypto.openssl_privatekey_pipe
  - module: community.crypto.openssl_publickey
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_aep_to_epg
short_description: Bind EPG to AEP (infra:RsFuncToEpg)
description:
- Bind EPG to AEP.
options:
  aep:
    description:
    - The name of the Attachable Access Entity Profile.
    type: str
    aliases: [ aep_name ]
  tenant:
    description:
    - The name of an existing tenant.
    type: str
    aliases: [ tenant_name ]
  ap:
    description:
    - Name of an existing application network profile, that will contain the EPGs.
    type: str
    aliases: [ app_profile, app_profile_name ]
  epg:
    description:
    - The name of the end point group.
    type: str
    aliases: [ epg_name ]
  encap:
    description:
    - The VLAN associated with this application EPG.
    type: int
    aliases: [ vlan, vlan_id, encap_id ]
  primary_encap:
    description:
    - The primary VLAN associated with this EPG
    type: int
    aliases: [ primary_vlan, primary_vlan_id, primary_encap_id ]
  interface_mode:
    description:
    - Determines how layer 2 tags will be read from and added to frames.
    - Values C(802.1p) and C(native) are identical.
    - Values C(access) and C(untagged) are identical.
    - Values C(regular), C(tagged) and C(trunk) are identical.
    - The APIC defaults to C(trunk) when unset during creation.
    type: str
    choices: [ 802.1p, access, native, regular, tagged, trunk, untagged ]
    aliases: [ mode, mode_name, interface_mode_name ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

author:
- Marcel Zehnder (@maercu)
"""

EXAMPLES = r"""
- name: Associate EPG with AEP
  cisco.aci.aci_aep_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    aep: aep1
    tenant: tenant1
    ap: ap1
    epg: epg1
    encap_id: 222
    interface_mode: access
    state: present
  delegate_to: localhost

- name: Associate EPG with AEP
  cisco.aci.aci_aep_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    aep: aep1
    tenant: tenant1
    ap: ap1
    epg: epg1
    encap_id: 222
    interface_mode: access
    state: absent
  delegate_to: localhost

- name: Get specific EPG with AEP association
  cisco.aci.aci_aep_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    aep: aep1
    tenant: tenant1
    ap: ap1
    epg: epg1
    encap_id: 222
    interface_mode: access
    state: query
  delegate_to: localhost
  register: query_result

- name: Get all EPG with AEP association
  cisco.aci.aci_aep_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""


from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec
from ansible.module_utils.basic import AnsibleModule


INTERFACE_MODE_MAPPING = {
    "802.1p": "native",
    "access": "untagged",
    "native": "native",
    "regular": "regular",
    "tagged": "regular",
    "trunk": "regular",
    "untagged": "untagged",
}


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        aep=dict(type="str", aliases=["aep_name"]),
        tenant=dict(type="str", aliases=["tenant_name"]),
        ap=dict(type="str", aliases=["app_profile", "app_profile_name"]),
        epg=dict(type="str", aliases=["epg_name"]),
        encap=dict(type="int", aliases=["vlan", "vlan_id", "encap_id"]),
        primary_encap=dict(type="int", aliases=["primary_vlan", "primary_vlan_id", "primary_encap_id"]),
        interface_mode=dict(
            type="str", choices=["802.1p", "access", "native", "regular", "tagged", "trunk", "untagged"], aliases=["mode_name", "mode", "interface_mode_name"]
        ),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["aep", "epg", "ap", "tenant"]],
            ["state", "present", ["interface_mode", "encap", "aep", "epg", "ap", "tenant"]],
        ],
    )

    aep = module.params.get("aep")
    tenant = module.params.get("tenant")
    ap = module.params.get("ap")
    epg = module.params.get("epg")
    encap = module.params.get("encap")
    primary_encap = module.params.get("primary_encap")
    interface_mode = module.params.get("interface_mode")
    state = module.params.get("state")

    if interface_mode is not None:
        interface_mode = INTERFACE_MODE_MAPPING[interface_mode]

    if encap is not None:
        encap = "vlan-{0}".format(encap)

    if primary_encap is not None:
        primary_encap = "vlan-{0}".format(primary_encap)

    epg_mo = None
    if tenant is not None and ap is not None and epg is not None:
        epg_mo = "uni/tn-{0}/ap-{1}/epg-{2}".format(tenant, ap, epg)

    aci = ACIModule(module)
    aci.construct_url(
        root_class=dict(aci_class="infraAttEntityP", aci_rn="infra/attentp-{0}".format(aep), module_object=aep, target_filter={"name": aep}),
        subclass_1=dict(aci_class="infraGeneric", aci_rn="gen-default", module_object="default", target_filter={"name": "default"}),
        subclass_2=dict(aci_class="infraRsFuncToEpg", aci_rn="rsfuncToEpg-[{0}]".format(epg_mo), module_object=epg_mo, target_filter={"tDn": epg_mo}),
    )

    aci.get_existing()

    if state == "present":
        # Post configuration on infraGeneric (subclass_1) level instead of on
        # infraRsFuncToEpg (subclass_2) level.
        # The reason being that the MO "gen-default" (of class infraGeneric) does not
        # exist until the first EPG to AEP association is created.
        aci.construct_url(
            root_class=dict(aci_class="infraAttEntityP", aci_rn="infra/attentp-{0}".format(aep), module_object=aep, target_filter={"name": aep}),
            subclass_1=dict(aci_class="infraGeneric", aci_rn="gen-default", module_object="default", target_filter={"name": "default"}),
            child_classes=["infraRsFuncToEpg"],
        )

        aci.get_existing()

        child_configs = [dict(infraRsFuncToEpg=dict(attributes=dict(encap=encap, primaryEncap=primary_encap, mode=interface_mode, tDn=epg_mo)))]

        aci.payload(aci_class="infraGeneric", class_config=dict(name="default"), child_configs=child_configs)

        aci.get_diff(aci_class="infraGeneric")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_certificate_management_policy_info
short_description: Gather information about Certificate Management Policies in Cisco Intersight
description:
  - Gather information about Certificate Management Policies in L(Cisco Intersight,https://intersight.com).
  - Information can be filtered by O(organization) and O(name).
  - If no filters are passed, all Certificate Management Policies will be returned.
  - Optionally retrieves associated certificates for each policy.
extends_documentation_fragment: intersight
options:
  organization:
    description:
      - The name of the organization the Certificate Management Policy belongs to.
    type: str
  name:
    description:
      - The name of the Certificate Management Policy to gather information from.
    type: str
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Fetch a specific Certificate Management Policy by name
  cisco.intersight.intersight_certificate_management_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "CertMgmt-Policy-01"

- name: Fetch all Certificate Management Policies in a specific Organization
  cisco.intersight.intersight_certificate_management_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"

- name: Fetch all Certificate Management Policies
  cisco.intersight.intersight_certificate_management_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: list
  elements: dict
  sample: [
    {
        "Name": "CertMgmt-Policy-01",
        "ObjectType": "certificatemanagement.Policy",
        "Moid": "1234567890abcdef12345678",
        "Description": "Certificate management policy with Root CA",
        "Organization": {
            "Name": "default",
            "Moid": "abcdef1234567890abcdef12",
            "ObjectType": "organization.Organization"
        },
        "Tags": [
            {
                "Key": "Environment",
                "Value": "Production"
            }
        ],
        "Certificates": [
            {
                "Certificate": {
                    "PemCertificate": "LS0tLS1CRUdJTiBDRVJUSUZJQ0FURS0tLS0t..."
                },
                "CertificateName": "RootCA-01",
                "ObjectType": "certificatemanagement.RootCaCertificate",
                "Enabled": true,
                "Moid": "cert1234567890abcdef12345"
            }
        ]
    },
    {
        "Name": "CertMgmt-Policy-02",
        "ObjectType": "certificatemanagement.Policy",
        "Moid": "5678901234abcdef56789012",
        "Description": "Certificate management policy with IMC certificate",
        "Organization": {
            "Name": "default",
            "Moid": "abcdef1234567890abcdef12",
            "ObjectType": "organization.Organization"
        },
        "Certificates": [
            {
                "Certificate": {
                    "PemCertificate": "LS0tLS1CRUdJTiBDRVJUSUZJQ0FURS0tLS0t..."
                },
                "ObjectType": "certificatemanagement.Imc",
                "Enabled": true,
                "IsPrivatekeySet": true,
                "Moid": "imccert1234567890abcdef"
            }
        ]
    }
  ]
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        organization=dict(type='str'),
        name=dict(type='str')
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    # Resource path used to fetch info
    resource_path = '/certificatemanagement/Policies'

    # Build query parameters
    query_params = intersight.set_query_params()

    # Get policies (certificates are included in the policy response)
    intersight.get_resource(
        resource_path=resource_path,
        query_params=query_params,
        return_list=True
    )

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2026, Rafal Kozlowski (@rkozlo) <rafalkozlowski07@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = r'''
---
module: clickhouse_named_collection

short_description: Creates, removes or modify a ClickHouse named collection using the clickhouse-driver Client interface

description:
  - Creates, remove or modify a ClickHouse named collection using the
    L(clickhouse-driver,https://clickhouse-driver.readthedocs.io/en/latest) Client interface.
  - The module can only work if O(login_user) will have necessary grants.
  - An existing named collection can be modified only if server is properly configured and user is allowed to
    L(view secrets,https://clickhouse.com/docs/sql-reference/statements/grant#displaysecretsinshowandselect).
    There is an option to rewrite whole secret, but then it will not be idempotent.
  - Module is supported only on version 25.8 or later.

attributes:
  check_mode:
    description: Supports check_mode.
    support: full
  idempotent:
    description:
      - At second run will not change anything.
      - If O(login_user) doesn't have proper privileges, the module can't determine if the collection differs.
      - O(rewrite) still will be idempotent if O(login_user) has privileges to display collection content.
    support: full

author:
  - Rafal Kozlowski (@rkozlo)

extends_documentation_fragment:
  - community.clickhouse.client_inst_opts

version_added: '2.2.0'

options:
  state:
    description:
      - Named collection state.
      - If C(present), will create or modify the named collection.
      - If C(absent), will drop the named collection if exists.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  name:
    description:
      - Named collection name to add or remove.
    type: str
    required: true
  rewrite:
    description:
      - Force to rewrite secret. It has only use when viewing secrets is disabled and user can't fetch secrets values.
      - If user can view secret values it will be fully idempotent. Will only alter collection if it differs.
    type: bool
    default: false
  collection:
    description:
      - Content of named collection.
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - Name of secret.
        type: str
        required: true
      value:
        description:
          - Content of secret.
        type: str
        required: true
  cluster:
    description:
      - Run the command on all cluster hosts.
      - If the cluster is not configured, the command will crash with an error.
    type: str
'''

EXAMPLES = r'''
- name: Create named collection with 2 secrets
  community.clickhouse.clickhouse_named_collection:
    name: test_col
    collection:
      - name: user
        value: alice
      - name: password
        value: test_pass

- name: Create named collection with forcing rewrite
  community.clickhouse.clickhouse_named_collection:
    name: test_col
    collection:
      - name: user
        value: alice
      - name: password
        value: test_pass
    rewrite: true

- name: Drop named collection
  community.clickhouse.clickhouse_named_collection:
    name: test_col
    state: absent
'''

RETURN = r'''
---
executed_statements:
  description:
    - Data-modifying executed statements.
  returned: on success
  type: list
  sample: ["CREATE NAMED COLLECTION `test_col` AS `user` = %(val_0)s, `password` = %(val_1)s"]

query_parameters:
  description:
    - Query parameters passed to query from executed_statements.
  returned: on succes
  type: list
  sample: [{"val_0": "VALUE_SPECIFIED_IN_NO_LOG_PARAMETER", "val_1": "VALUE_SPECIFIED_IN_NO_LOG_PARAMETER"}]
'''

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.clickhouse.plugins.module_utils.clickhouse import (
    check_clickhouse_driver,
    client_common_argument_spec,
    connect_to_db_via_client,
    execute_query,
    get_main_conn_kwargs,
    get_server_version,
    get_on_cluster_clause,
    validate_identifier,
)

executed_statements = []
query_parameters = []


class ClickHouseNamedCollection:
    def __init__(self, module, client, name):
        self.module = module
        self.client = client
        validate_identifier(module, name, "collection name")
        self.name = name
        self._exists = None
        self._source = None
        self._current = None
        self._query_parameters = {}

    @property
    def exists(self):
        self._load()
        return self._exists

    @property
    def source(self):
        self._load()
        return self._source

    @property
    def current(self):
        self._load()
        return self._current

    def _load(self):
        """Fetch info about passed collection"""
        if self._exists is not None:
            return

        query = f"SELECT collection, source FROM system.named_collections WHERE name = '{self.name}'"
        result = execute_query(self.module, self.client, query)

        if not result:
            self._exists = False
            self._source = None
            self._current = None
            return

        collection_data, source = result[0]
        self._exists = True
        self._source = source

        if source == 'SQL':
            self._current = self._normalize_current_collection_data(collection_data)
        else:
            self.module.fail_json(
                msg=f"Passed named collection isn't sourced by SQL. Got: {source}"
            )

    def _normalize_current_collection_data(self, collection):
        """Normalize raw db output into normalized form used in module options."""
        """At this moment only support value field."""

        result = {}
        for key, value in collection.items():
            result[key] = {'value': value}
            self.module.no_log_values.add(value)
        return result

    def _check_if_hidden(self):
        """Check normalized collection if hasn't hidden fields."""
        """At this moment it can't be hidden in single key so we return on single element as whole."""
        for key in self._current.values():
            if key['value'] == '[HIDDEN]':
                return True
        return False

    def _build_list_collection(self, collection):
        parts = []
        i = 0
        for name, content in collection.items():
            validate_identifier(self.module, name, "collection key")

            for value in content.values():
                parts.append(f"`{name}` = %(val_{i})s")
                self._query_parameters[f'val_{i}'] = value
                i += 1
        return parts

    def _should_skip_alter(self, collection, rewrite):
        """Wheter preceed alter or not."""
        hidden = self._check_if_hidden()

        # Hidden means we can't determinie difference in key values.
        if hidden and not rewrite:
            return True

        if self.current == collection:
            return True
        return False

    def _build_alter_query(self, collection, cluster):
        parts = self._build_list_collection(collection)
        query = f"ALTER NAMED COLLECTION `{self.name}`"
        query += get_on_cluster_clause(self.module, cluster)

        query += " SET " + ", ".join(parts)
        return query

    def create(self, collection, cluster):
        if not collection:
            self.module.fail_json(msg="Collection not passed")
        query = f"CREATE NAMED COLLECTION `{self.name}`"
        query += get_on_cluster_clause(self.module, cluster)

        query += " AS "
        parts = self._build_list_collection(collection)
        query += ", ".join(parts)
        execute_kwargs = {'params': self._query_parameters}
        executed_statements.append(query)

        query_parameters.append(self._query_parameters.copy())

        if not self.module.check_mode:
            execute_query(self.module, self.client, query, execute_kwargs)
        return True

    def drop(self, cluster):
        query = f"DROP NAMED COLLECTION `{self.name}`"

        query += get_on_cluster_clause(self.module, cluster)

        executed_statements.append(query)

        query_parameters.append(self._query_parameters.copy())
        execute_kwargs = {'params': self._query_parameters}
        if not self.module.check_mode:
            execute_query(self.module, self.client, query, execute_kwargs)

        return True

    def alter(self, collection, cluster, rewrite):
        if not collection:
            self.module.fail_json(msg="Collection not passed")
        if self._should_skip_alter(collection, rewrite):
            return False
        query = self._build_alter_query(collection, cluster)
        executed_statements.append(query)
        query_parameters.append(self._query_parameters.copy())
        execute_kwargs = {'params': self._query_parameters}
        if not self.module.check_mode:
            execute_query(self.module, self.client, query, execute_kwargs)
        return True

    def _normalize_collection_input(self, collection_list):
        """Convert list of {name, value} to dict of {name: {value}}"""
        if not collection_list:
            return {}

        normalized = {}
        for item in collection_list:
            name = item['name']
            value = item['value']

            # Handle duplicate names (last one wins or fail)
            if name in normalized:
                self.module.warn(f"Duplicate parameter '{name}', using last value")

            normalized[name] = {'value': value}

        return normalized


def main():
    # Set up arguments.
    # If there are common arguments shared across several modules,
    # create the common_argument_spec() function under plugins/module_utils/*
    # and invoke here to return a dict with those arguments
    argument_spec = client_common_argument_spec()
    argument_spec.update(
        state=dict(type="str", choices=["present", "absent"], default="present"),
        name=dict(type="str", required=True),
        cluster=dict(type='str', default=None),
        collection=dict(
            type='list',
            elements='dict',
            options=dict(
                name=dict(type='str', required=True),
                value=dict(type='str', required=True, no_log=True)
            )
        ),
        rewrite=dict(type='bool', default=False)
    )

    # Conditional logic
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ('state', 'present', ['collection'], True)
        ],
    )

    # Assign passed options to variables
    client_kwargs = module.params["client_kwargs"]
    # The reason why these arguments are separate from client_kwargs
    # is that we need to protect some sensitive data like passwords passed
    # to the module from logging (see the arguments above with no_log=True);
    # Such data must be passed as module arguments (not nested deep in values).
    main_conn_kwargs = get_main_conn_kwargs(module)
    state = module.params["state"]
    name = module.params["name"]
    cluster = module.params["cluster"]
    collection = module.params["collection"]
    rewrite = module.params["rewrite"]

    # Will fail if no driver informing the user
    check_clickhouse_driver(module)

    # Connect to DB
    client = connect_to_db_via_client(module, main_conn_kwargs, client_kwargs)
    server_version = get_server_version(module, client)

    if server_version['year'] < 25 or (server_version['year'] == 25 and server_version['feature'] < 8):
        module.fail_json(msg="Server version not supported. Require 25.8 or later.")

    # Do the job
    changed = False
    named_collection = ClickHouseNamedCollection(module, client, name)
    normalized_collection = named_collection._normalize_collection_input(collection)
    if state == "present":
        if not named_collection.exists:
            changed = named_collection.create(normalized_collection, cluster)
        else:
            changed = named_collection.alter(normalized_collection, cluster, rewrite)
    else:
        # If state is absent
        if named_collection.exists:
            changed = named_collection.drop(cluster)

    # Close connection
    client.disconnect_connection()

    # Users will get this in JSON output after execution
    module.exit_json(changed=changed, executed_statements=executed_statements, query_parameters=query_parameters)


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import typing as t
from urllib.error import HTTPError

from ansible.module_utils.common.text.converters import to_native
from ansible.module_utils.urls import (
    ConnectionError,  # noqa: A004
    NoSSLError,
    fetch_url,
    open_url,
)

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    HTTPMethod = t.Literal["DELETE", "GET", "HEAD", "POST", "PUT"]  # pragma: no cover


class NetworkError(Exception):
    pass


class HTTPHelper(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def fetch_url(
        self,
        url: str,
        *,
        method: HTTPMethod = "GET",
        headers: dict[str, str] | None = None,
        data: bytes | None = None,
        timeout: int | None = None,
    ) -> tuple[bytes | None, dict[str, t.Any]]:
        """
        Execute a HTTP request and return a tuple (response_content, info).

        In case of errors, either raise NetworkError or terminate the program (for modules only!).
        """


class ModuleHTTPHelper(HTTPHelper):
    def __init__(
        self,
        module: AnsibleModule,
    ) -> None:
        self.module: AnsibleModule = module

    def fetch_url(
        self,
        url: str,
        *,
        method: HTTPMethod = "GET",
        headers: dict[str, str] | None = None,
        data: bytes | None = None,
        timeout: int | None = None,
    ) -> tuple[bytes | None, dict[str, t.Any]]:
        response, info = fetch_url(
            self.module, url, method=method, headers=headers, data=data, timeout=timeout
        )
        try:
            # read() from a closed response returns ''
            if response.closed:
                raise TypeError
            content = response.read()
        except (AttributeError, TypeError):
            content = info.pop("body", None)
        return content, info


class OpenURLHelper(HTTPHelper):
    def fetch_url(
        self,
        url: str,
        *,
        method: HTTPMethod = "GET",
        headers: dict[str, str] | None = None,
        data: bytes | None = None,
        timeout: int | None = None,
    ) -> tuple[bytes | None, dict[str, t.Any]]:
        info = {}
        try:
            req = open_url(
                url, method=method, headers=headers, data=data, timeout=timeout
            )
            result = req.read()
            info.update({k.lower(): v for k, v in req.info().items()})
            info["status"] = req.code
            info["url"] = req.geturl()
            req.close()
        except HTTPError as e:
            try:
                result = e.read()
            except AttributeError:
                result = ""
            try:
                info.update({k.lower(): v for k, v in e.info().items()})
            except Exception:  # pragma: no cover
                pass  # pragma: no cover
            info["status"] = e.code
        except NoSSLError as e:
            raise NetworkError(f"Cannot connect via SSL: {to_native(e)}") from e
        except (ConnectionError, ValueError) as e:
            raise NetworkError(f"Connection error: {to_native(e)}") from e

        return result, info



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, René Moser <mail@renemoser.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: server_group
short_description: Manages server groups on the cloudscale.ch IaaS service
description:
  - Create, update and remove server groups.
author:
  - René Moser (@resmo)
  - Denis Krienbühl (@href)
version_added: "1.0.0"
options:
  name:
    description:
      - Name of the server group.
      - Either I(name) or I(uuid) is required. These options are mutually exclusive.
    type: str
  uuid:
    description:
      - UUID of the server group.
      - Either I(name) or I(uuid) is required. These options are mutually exclusive.
    type: str
  type:
    description:
      - Type of the server group.
    default: anti-affinity
    type: str
  zone:
    description:
      - Zone slug of the server group (e.g. C(lpg1) or C(rma1)).
    type: str
  state:
    description:
      - State of the server group.
    choices: [ present, absent ]
    default: present
    type: str
  tags:
    description:
      - Tags assosiated with the server groups. Set this to C({}) to clear any tags.
    type: dict
extends_documentation_fragment: cloudscale_ch.cloud.api_parameters
'''

EXAMPLES = '''
---
- name: Ensure server group exists
  cloudscale_ch.cloud.server_group:
    name: my-name
    type: anti-affinity
    api_token: xxxxxx

- name: Ensure server group in a specific zone
  cloudscale_ch.cloud.server_group:
    name: my-rma-group
    type: anti-affinity
    zone: lpg1
    api_token: xxxxxx

- name: Ensure a server group is absent
  cloudscale_ch.cloud.server_group:
    name: my-name
    state: absent
    api_token: xxxxxx
'''

RETURN = '''
---
href:
  description: API URL to get details about this server group
  returned: if available
  type: str
  sample: https://api.cloudscale.ch/v1/server-group/cfde831a-4e87-4a75-960f-89b0148aa2cc
uuid:
  description: The unique identifier for this server
  returned: always
  type: str
  sample: cfde831a-4e87-4a75-960f-89b0148aa2cc
name:
  description: The display name of the server group
  returned: always
  type: str
  sample: load balancers
type:
  description: The type the server group
  returned: if available
  type: str
  sample: anti-affinity
zone:
  description: The zone of the server group
  returned: success
  type: dict
  sample: { 'slug': 'rma1' }
servers:
  description: A list of servers that are part of the server group.
  returned: if available
  type: list
  sample: []
state:
  description: State of the server group.
  returned: always
  type: str
  sample: present
tags:
  description: Tags assosiated with the server group.
  returned: success
  type: dict
  sample: { 'project': 'my project' }
'''

from ansible.module_utils.basic import AnsibleModule
from ..module_utils.api import (
    AnsibleCloudscaleBase,
    cloudscale_argument_spec,
)


def main():
    argument_spec = cloudscale_argument_spec()
    argument_spec.update(dict(
        name=dict(type='str'),
        uuid=dict(type='str'),
        type=dict(type='str', default='anti-affinity'),
        zone=dict(type='str'),
        tags=dict(type='dict'),
        state=dict(default='present', choices=['absent', 'present']),
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_one_of=(('name', 'uuid'),),
        required_if=(('state', 'present', ('name',),),),
        supports_check_mode=True,
    )

    cloudscale_server_group = AnsibleCloudscaleBase(
        module,
        resource_name='server-groups',
        resource_create_param_keys=[
            'name',
            'type',
            'zone',
            'tags',
        ],
        resource_update_param_keys=[
            'name',
            'tags',
        ],
    )
    cloudscale_server_group.query_constraint_keys = [
        'zone',
    ]

    if module.params['state'] == 'absent':
        result = cloudscale_server_group.absent()
    else:
        result = cloudscale_server_group.present()
    module.exit_json(**result)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The eos lag_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

import re

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.arista.eos.plugins.module_utils.network.eos.argspec.lag_interfaces.lag_interfaces import (
    Lag_interfacesArgs,
)


class Lag_interfacesFacts(object):
    """The eos lag_interfaces fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Lag_interfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_device_data(self, connection):
        return connection.get("show running-config | section ^interface")

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for lag_interfaces
        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected configuration
        :rtype: dictionary
        :returns: facts
        """
        if not data:
            data = self.get_device_data(connection)

        # split the config into instances of the resource
        resource_delim = "interface"
        find_pattern = r"(?:^|\n)%s.*?(?=(?:^|\n)%s|$)" % (
            resource_delim,
            resource_delim,
        )
        resources = [p.strip() for p in re.findall(find_pattern, data, re.DOTALL)]

        objs = {}
        for resource in resources:
            if resource:
                obj = self.render_config(self.generated_spec, resource)
                if obj:
                    group_name = obj["name"]
                    if group_name in objs and "members" in obj:
                        config = objs[group_name]
                        if "members" not in config:
                            config["members"] = []
                        objs[group_name]["members"].extend(obj["members"])
                    else:
                        objs[group_name] = obj
        objs = list(objs.values())
        facts = {"lag_interfaces": []}
        if objs:
            params = utils.validate_config(
                self.argument_spec,
                {"config": objs},
            )
            facts["lag_interfaces"] = [utils.remove_empties(cfg) for cfg in params["config"]]
        ansible_facts["ansible_network_resources"].update(facts)
        return ansible_facts

    def render_config(self, spec, conf):
        """
        Render config as dictionary structure and delete keys
          from spec for null values

        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)
        interface_name = utils.parse_conf_arg(conf, "interface")
        if interface_name.startswith("Port-Channel"):
            config["name"] = interface_name
            return utils.remove_empties(config)

        interface = {"member": interface_name}
        match = re.match(
            r".*channel-group (\d+) mode (\S+)",
            conf,
            re.MULTILINE | re.DOTALL,
        )
        if match:
            config["name"], interface["mode"] = match.groups()
            config["name"] = "Port-Channel" + config["name"]
            config["members"] = [interface]

        return utils.remove_empties(config)



#!/usr/bin/python

# Copyright (c) 2014, Kevin Carter <kevin.carter@rackspace.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: lxc_container
short_description: Manage LXC Containers
description:
  - Management of LXC containers.
author: "Kevin Carter (@cloudnull)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of a container.
    type: str
    required: true
  backing_store:
    choices:
      - dir
      - lvm
      - loop
      - btrfs
      - overlayfs
      - zfs
    description:
      - Backend storage type for the container.
    type: str
    default: dir
  template:
    description:
      - Name of the template to use within an LXC create.
    type: str
    default: ubuntu
  template_options:
    description:
      - Template options when building the container.
    type: str
  config:
    description:
      - Path to the LXC configuration file.
    type: path
  lv_name:
    description:
      - Name of the logical volume, defaults to the container name.
      - If not specified, it defaults to E(CONTAINER_NAME).
    type: str
  vg_name:
    description:
      - If backend store is lvm, specify the name of the volume group.
    type: str
    default: lxc
  thinpool:
    description:
      - Use LVM thin pool called TP.
    type: str
  fs_type:
    description:
      - Create fstype TYPE.
    type: str
    default: ext4
  fs_size:
    description:
      - File system Size.
    type: str
    default: 5G
  directory:
    description:
      - Place rootfs directory under DIR.
    type: path
  zfs_root:
    description:
      - Create zfs under given zfsroot.
    type: str
  container_command:
    description:
      - Run a command within a container.
      - The O(container_command) can be used with any state except V(absent). If used with state V(stopped) the container
        is V(started), the command executed, and then the container V(stopped) again. Likewise if O(state=stopped) and the
        container does not exist it is first created, V(started), the command executed, and then V(stopped). If you use a
        C(|) in the variable you can use common script formatting within the variable itself. The O(container_command) option
        always execute as C(bash). When using O(container_command), a log file is created in the C(/tmp/) directory which
        contains both RV(ignore:stdout) and RV(ignore:stderr) of any command executed.
    type: str
  lxc_path:
    description:
      - Place container under E(PATH).
    type: path
  container_log:
    description:
      - Enable a container log for host actions to the container.
    type: bool
    default: false
  container_log_level:
    choices:
      - Info
      - info
      - INFO
      - Error
      - error
      - ERROR
      - Debug
      - debug
      - DEBUG
    description:
      - Set the log level for a container where O(container_log) was set.
    type: str
    default: INFO
  clone_name:
    description:
      - Name of the new cloned server.
      - This is only used when state is clone.
    type: str
  clone_snapshot:
    description:
      - Create a snapshot a container when cloning.
      - This is not supported by all container storage backends.
      - Enabling this may fail if the backing store does not support snapshots.
    type: bool
    default: false
  archive:
    description:
      - When set to V(true) the system attempts to create a compressed tarball of the running container. The O(archive) option
        supports LVM backed containers and creates a snapshot of the running container when creating the archive.
    type: bool
    default: false
  archive_path:
    description:
      - Path the save the archived container.
      - If the path does not exist the archive method attempts to create it.
    type: path
  archive_compression:
    choices:
      - gzip
      - bzip2
      - none
    description:
      - Type of compression to use when creating an archive of a running container.
    type: str
    default: gzip
  state:
    choices:
      - started
      - stopped
      - restarted
      - absent
      - frozen
      - clone
    description:
      - Define the state of a container.
      - If you clone a container using O(clone_name) the newly cloned container created in a stopped state.
      - The running container is stopped while the clone operation is happening and upon completion of the clone the original
        container state is restored.
    type: str
    default: started
  container_config:
    description:
      - A list of C(key=value) options to use when configuring a container.
    type: list
    elements: str
requirements:
  - 'lxc >= 2.0 # OS package'
  - 'python3 >= 3.5 # OS Package'
  - 'python3-lxc # OS Package'
notes:
  - Containers must have a unique name. If you attempt to create a container with a name that already exists in the users
    namespace the module simply returns as "unchanged".
  - If your distro does not have a package for C(python3-lxc), which is a requirement for this module, it can be installed
    from source at U(https://github.com/lxc/python3-lxc) or installed using C(pip install lxc).
"""

EXAMPLES = r"""
- name: Create a started container
  community.general.lxc_container:
    name: test-container-started
    container_log: true
    template: ubuntu
    state: started
    template_options: --release trusty

- name: Create a stopped container
  community.general.lxc_container:
    name: test-container-stopped
    container_log: true
    template: ubuntu
    state: stopped
    template_options: --release trusty

- name: Create a frozen container
  community.general.lxc_container:
    name: test-container-frozen
    container_log: true
    template: ubuntu
    state: frozen
    template_options: --release trusty
    container_command: |
      echo 'hello world.' | tee /opt/started-frozen

# Create filesystem container, configure it, and archive it, and start it.
- name: Create filesystem container
  community.general.lxc_container:
    name: test-container-config
    backing_store: dir
    container_log: true
    template: ubuntu
    state: started
    archive: true
    archive_compression: none
    container_config:
      - "lxc.aa_profile=unconfined"
      - "lxc.cgroup.devices.allow=a *:* rmw"
    template_options: --release trusty

# Create an lvm container, run a complex command in it, add additional
# configuration to it, create an archive of it, and finally leave the container
# in a frozen state. The container archive will be compressed using bzip2
- name: Create a frozen lvm container
  community.general.lxc_container:
    name: test-container-lvm
    container_log: true
    template: ubuntu
    state: frozen
    backing_store: lvm
    template_options: --release trusty
    container_command: |
      apt-get update
      apt-get install -y vim lxc-dev
      echo 'hello world.' | tee /opt/started
      if [[ -f "/opt/started" ]]; then
          echo 'hello world.' | tee /opt/found-started
      fi
    container_config:
      - "lxc.aa_profile=unconfined"
      - "lxc.cgroup.devices.allow=a *:* rmw"
    archive: true
    archive_compression: bzip2
  register: lvm_container_info

- name: Debug info on container "test-container-lvm"
  ansible.builtin.debug:
    var: lvm_container_info

- name: Run a command in a container and ensure it is in a "stopped" state.
  community.general.lxc_container:
    name: test-container-started
    state: stopped
    container_command: |
      echo 'hello world.' | tee /opt/stopped

- name: Run a command in a container and ensure it is in a "frozen" state.
  community.general.lxc_container:
    name: test-container-stopped
    state: frozen
    container_command: |
      echo 'hello world.' | tee /opt/frozen

- name: Start a container
  community.general.lxc_container:
    name: test-container-stopped
    state: started

- name: Run a command in a container and then restart it
  community.general.lxc_container:
    name: test-container-started
    state: restarted
    container_command: |
      echo 'hello world.' | tee /opt/restarted

- name: Run a complex command within a "running" container
  community.general.lxc_container:
    name: test-container-started
    container_command: |
      apt-get update
      apt-get install -y curl wget vim apache2
      echo 'hello world.' | tee /opt/started
      if [[ -f "/opt/started" ]]; then
          echo 'hello world.' | tee /opt/found-started
      fi

# Create an archive of an existing container, save the archive to a defined
# path and then destroy it.
- name: Archive container
  community.general.lxc_container:
    name: test-container-started
    state: absent
    archive: true
    archive_path: /opt/archives

# Create a container using overlayfs, create an archive of it, create a
# snapshot clone of the container and and finally leave the container
# in a frozen state. The container archive will be compressed using gzip.
- name: Create an overlayfs container archive and clone it
  community.general.lxc_container:
    name: test-container-overlayfs
    container_log: true
    template: ubuntu
    state: started
    backing_store: overlayfs
    template_options: --release trusty
    clone_snapshot: true
    clone_name: test-container-overlayfs-clone-snapshot
    archive: true
    archive_compression: gzip
  register: clone_container_info

- name: Debug info on container "test-container"
  ansible.builtin.debug:
    var: clone_container_info

- name: Clone a container using snapshot
  community.general.lxc_container:
    name: test-container-overlayfs-clone-snapshot
    backing_store: overlayfs
    clone_name: test-container-overlayfs-clone-snapshot2
    clone_snapshot: true

- name: Create a new container and clone it
  community.general.lxc_container:
    name: test-container-new-archive
    backing_store: dir
    clone_name: test-container-new-archive-clone

- name: Archive and clone a container then destroy it
  community.general.lxc_container:
    name: test-container-new-archive
    state: absent
    clone_name: test-container-new-archive-destroyed-clone
    archive: true
    archive_compression: gzip

- name: Start a cloned container.
  community.general.lxc_container:
    name: test-container-new-archive-destroyed-clone
    state: started

- name: Destroy a container
  community.general.lxc_container:
    name: '{{ item }}'
    state: absent
  with_items:
    - test-container-stopped
    - test-container-started
    - test-container-frozen
    - test-container-lvm
    - test-container-config
    - test-container-overlayfs
    - test-container-overlayfs-clone
    - test-container-overlayfs-clone-snapshot
    - test-container-overlayfs-clone-snapshot2
    - test-container-new-archive
    - test-container-new-archive-clone
    - test-container-new-archive-destroyed-clone
"""

RETURN = r"""
lxc_container:
  description: Container information.
  returned: success
  type: complex
  contains:
    name:
      description: Name of the LXC container.
      returned: success
      type: str
      sample: test_host
    init_pid:
      description: Pid of the LXC init process.
      returned: success
      type: int
      sample: 19786
    interfaces:
      description: List of the container's network interfaces.
      returned: success
      type: list
      sample: ["eth0", "lo"]
    ips:
      description: List of IPs.
      returned: success
      type: list
      sample: ["10.0.3.3"]
    state:
      description: Resulting state of the container.
      returned: success
      type: str
      sample: "running"
    archive:
      description: Resulting state of the container.
      returned: success, when archive is true
      type: str
      sample: "/tmp/test-container-config.tar"
    clone:
      description: If the container was cloned.
      returned: success, when clone_name is specified
      type: bool
      sample: true
"""

import os
import re
import shlex
import tempfile
import time

try:
    import lxc
except ImportError:
    HAS_LXC = False
else:
    HAS_LXC = True

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.module_utils.parsing.convert_bool import BOOLEANS_FALSE

from ansible_collections.community.general.plugins.module_utils._lvm import (
    lvcreate_runner,
    lvremove_runner,
    lvs_runner,
    vgs_runner,
)
from ansible_collections.community.general.plugins.module_utils._lxc import create_script

# LXC_COMPRESSION_MAP is a map of available compression types when creating
# an archive of a container.
LXC_COMPRESSION_MAP = {
    "gzip": {"extension": "tar.tgz", "argument": "-czf"},
    "bzip2": {"extension": "tar.bz2", "argument": "-cjf"},
    "none": {"extension": "tar", "argument": "-cf"},
}


# LXC_COMMAND_MAP is a map of variables that are available to a method based
# on the state the container is in.
LXC_COMMAND_MAP = {
    "create": {
        "variables": {
            "config": "--config",
            "template": "--template",
            "backing_store": "--bdev",
            "lxc_path": "--lxcpath",
            "lv_name": "--lvname",
            "vg_name": "--vgname",
            "thinpool": "--thinpool",
            "fs_type": "--fstype",
            "fs_size": "--fssize",
            "directory": "--dir",
            "zfs_root": "--zfsroot",
        }
    },
    "clone": {
        "variables-lxc-copy": {
            "backing_store": "--backingstorage",
            "lxc_path": "--lxcpath",
            "fs_size": "--fssize",
            "name": "--name",
            "clone_name": "--newname",
        },
        # lxc-clone is deprecated in favor of lxc-copy
        "variables-lxc-clone": {
            "backing_store": "--backingstore",
            "lxc_path": "--lxcpath",
            "fs_size": "--fssize",
            "name": "--orig",
            "clone_name": "--new",
        },
    },
}


# LXC_BACKING_STORE is a map of available storage backends and options that
# are incompatible with the given storage backend.
LXC_BACKING_STORE = {
    "dir": ["lv_name", "vg_name", "fs_type", "fs_size", "thinpool"],
    "lvm": ["zfs_root"],
    "btrfs": ["lv_name", "vg_name", "thinpool", "zfs_root", "fs_type", "fs_size"],
    "loop": ["lv_name", "vg_name", "thinpool", "zfs_root"],
    "overlayfs": ["lv_name", "vg_name", "fs_type", "fs_size", "thinpool", "zfs_root"],
    "zfs": ["lv_name", "vg_name", "fs_type", "fs_size", "thinpool"],
}


# LXC_LOGGING_LEVELS is a map of available log levels
LXC_LOGGING_LEVELS = {
    "INFO": ["info", "INFO", "Info"],
    "ERROR": ["error", "ERROR", "Error"],
    "DEBUG": ["debug", "DEBUG", "Debug"],
}


# LXC_ANSIBLE_STATES is a map of states that contain values of methods used
# when a particular state is evoked.
LXC_ANSIBLE_STATES = {
    "started": "_started",
    "stopped": "_stopped",
    "restarted": "_restarted",
    "absent": "_destroyed",
    "frozen": "_frozen",
    "clone": "_clone",
}


class LxcContainerManagement:
    def __init__(self, module):
        """Management of LXC containers via Ansible.

        :param module: Processed Ansible Module.
        :type module: ``object``
        """
        self.module = module
        self.state = self.module.params["state"]
        self.state_change = False
        self.lxc_vg = None
        self.lxc_path = self.module.params["lxc_path"]
        self.container_name = self.module.params["name"]
        self.container = self.get_container_bind()
        self.archive_info = None
        self.clone_info = None
        self.run_info = []

    def get_container_bind(self):
        return lxc.Container(name=self.container_name)

    @staticmethod
    def _roundup(num):
        """Return a rounded floating point number.

        :param num: Number to round up.
        :type: ``float``
        :returns: Rounded up number.
        :rtype: ``int``
        """
        num, part = str(num).split(".")
        num = int(num)
        if int(part) != 0:
            num += 1
        return num

    @staticmethod
    def _container_exists(container_name, lxc_path=None):
        """Check if a container exists.

        :param container_name: Name of the container.
        :type: ``str``
        :returns: True or False if the container is found.
        :rtype: ``bol``
        """
        return any(c == container_name for c in lxc.list_containers(config_path=lxc_path))

    @staticmethod
    def _add_variables(variables_dict, build_command):
        """Return a command list with all found options.

        :param variables_dict: Pre-parsed optional variables used from a
                               seed command.
        :type variables_dict: ``dict``
        :param build_command: Command to run.
        :type build_command: ``list``
        :returns: list of command options.
        :rtype: ``list``
        """

        for key, value in variables_dict.items():
            build_command.append(str(key))
            build_command.append(str(value))
        return build_command

    def _get_vars(self, variables):
        """Return a dict of all variables as found within the module.

        :param variables: Hash of all variables to find.
        :type variables: ``dict``
        """

        # Remove incompatible storage backend options.
        variables = variables.copy()
        for v in LXC_BACKING_STORE[self.module.params["backing_store"]]:
            variables.pop(v, None)

        false_values = BOOLEANS_FALSE.union([None, ""])
        result = {v: self.module.params[k] for k, v in variables.items() if self.module.params[k] not in false_values}
        return result

    def _config(self):
        """Configure an LXC container.

        Write new configuration values to the lxc config file. This will
        stop the container if it is running write the new options and then
        restart the container upon completion.
        """

        _container_config = self.module.params["container_config"]
        if not _container_config:
            return False

        container_config_file = self.container.config_file_name
        with open(container_config_file, "rb") as f:
            container_config = to_text(f.read(), errors="surrogate_or_strict").splitlines(True)

        parsed_options = [i.split("=", 1) for i in _container_config]
        config_change = False
        for key, value in parsed_options:
            key = key.strip()
            value = value.strip()
            new_entry = f"{key} = {value}\n"
            keyre = re.compile(rf"{key}(\s+)?=")
            for option_line in container_config:
                # Look for key in config
                if keyre.match(option_line):
                    dummy, _value = option_line.split("=", 1)
                    config_value = " ".join(_value.split())
                    line_index = container_config.index(option_line)
                    # If the sanitized values don't match replace them
                    if value != config_value:
                        line_index += 1
                        if new_entry not in container_config:
                            config_change = True
                            container_config.insert(line_index, new_entry)
                    # Break the flow as values are written or not at this point
                    break
            else:
                config_change = True
                container_config.append(new_entry)

        # If the config changed restart the container.
        if config_change:
            container_state = self._get_state()
            if container_state != "stopped":
                self.container.stop()

            with open(container_config_file, "wb") as f:
                f.writelines([to_bytes(line, errors="surrogate_or_strict") for line in container_config])

            self.state_change = True
            if container_state == "running":
                self._container_startup()
            elif container_state == "frozen":
                self._container_startup()
                self.container.freeze()

    def _container_create_clone(self):
        """Clone a new LXC container from an existing container.

        This method will clone an existing container to a new container using
        the `clone_name` variable as the new container name. The method will
        create a container if the container `name` does not exist.

        Note that cloning a container will ensure that the original container
        is "stopped" before the clone can be done. Because this operation can
        require a state change the method will return the original container
        to its prior state upon completion of the clone.

        Once the clone is complete the new container will be left in a stopped
        state.
        """

        # Ensure that the state of the original container is stopped
        container_state = self._get_state()
        if container_state != "stopped":
            self.state_change = True
            self.container.stop()

        # lxc-clone is deprecated in favor of lxc-copy
        clone_vars = "variables-lxc-copy"
        clone_cmd = self.module.get_bin_path("lxc-copy")
        if not clone_cmd:
            clone_vars = "variables-lxc-clone"
            clone_cmd = self.module.get_bin_path("lxc-clone", True)

        build_command = [
            clone_cmd,
        ]

        build_command = self._add_variables(
            variables_dict=self._get_vars(variables=LXC_COMMAND_MAP["clone"][clone_vars]), build_command=build_command
        )

        # Load logging for the instance when creating it.
        if self.module.params["clone_snapshot"]:
            build_command.append("--snapshot")
        # Check for backing_store == overlayfs if so force the use of snapshot
        # If overlay fs is used and snapshot is unset the clone command will
        # fail with an unsupported type.
        elif self.module.params["backing_store"] == "overlayfs":
            build_command.append("--snapshot")

        rc, return_data, err = self.module.run_command(build_command)
        if rc != 0:
            message = f"Failed executing {os.path.basename(clone_cmd)}."
            self.failure(err=err, rc=rc, msg=message, command=" ".join(build_command))
        else:
            self.state_change = True
            # Restore the original state of the origin container if it was
            # not in a stopped state.
            if container_state == "running":
                self.container.start()
            elif container_state == "frozen":
                self.container.start()
                self.container.freeze()

        return True

    def _create(self):
        """Create a new LXC container.

        This method will build and execute a shell command to build the
        container. It would have been nice to simply use the lxc python library
        however at the time this was written the python library, in both py2
        and py3 didn't support some of the more advanced container create
        processes. These missing processes mainly revolve around backing
        LXC containers with block devices.
        """

        build_command = [self.module.get_bin_path("lxc-create", True), "--name", self.container_name, "--quiet"]

        build_command = self._add_variables(
            variables_dict=self._get_vars(variables=LXC_COMMAND_MAP["create"]["variables"]), build_command=build_command
        )

        # Load logging for the instance when creating it.
        if self.module.params["container_log"]:
            # Set the logging path to the /var/log/lxc if uid is root. else
            # set it to the home folder of the user executing.
            try:
                if os.getuid() != 0:
                    log_path = os.getenv("HOME")
                else:
                    if not os.path.isdir("/var/log/lxc/"):
                        os.makedirs("/var/log/lxc/")
                    log_path = "/var/log/lxc/"
            except OSError:
                log_path = os.getenv("HOME")

            build_command.extend(
                [
                    "--logfile",
                    os.path.join(log_path, f"lxc-{self.container_name}.log"),
                    "--logpriority",
                    self.module.params.get("container_log_level").upper(),
                ]
            )

        # Add the template commands to the end of the command if there are any
        template_options = self.module.params["template_options"]
        if template_options:
            build_command.append("--")
            build_command += shlex.split(template_options)

        rc, return_data, err = self.module.run_command(build_command)
        if rc != 0:
            message = "Failed executing lxc-create."
            self.failure(err=err, rc=rc, msg=message, command=" ".join(build_command))
        else:
            self.state_change = True

    def _container_data(self):
        """Returns a dict of container information.

        :returns: container data
        :rtype: ``dict``
        """

        return {
            "interfaces": self.container.get_interfaces(),
            "ips": self.container.get_ips(),
            "state": self._get_state(),
            "init_pid": int(self.container.init_pid),
            "name": self.container_name,
        }

    def _unfreeze(self):
        """Unfreeze a container.

        :returns: True or False based on if the container was unfrozen.
        :rtype: ``bol``
        """

        unfreeze = self.container.unfreeze()
        if unfreeze:
            self.state_change = True
        return unfreeze

    def _get_state(self):
        """Return the state of a container.

        If the container is not found the state returned is "absent"

        :returns: state of a container as a lower case string.
        :rtype: ``str``
        """

        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            return str(self.container.state).lower()
        return "absent"

    def _execute_command(self):
        """Execute a shell command."""

        container_command = self.module.params["container_command"]
        if container_command:
            container_state = self._get_state()
            if container_state == "frozen":
                self._unfreeze()
            elif container_state == "stopped":
                self._container_startup()

            self.container.attach_wait(create_script, (container_command, self.module))
            self.state_change = True

    def _container_startup(self, timeout=60):
        """Ensure a container is started.

        :param timeout: Time before the destroy operation is abandoned.
        :type timeout: ``int``
        """

        self.container = self.get_container_bind()
        for dummy in range(timeout):
            if self._get_state() == "running":
                return True

            self.container.start()
            self.state_change = True
            # post startup sleep for 1 second.
            time.sleep(1)
        self.failure(
            lxc_container=self._container_data(),
            error=f"Failed to start container [ {self.container_name} ]",
            rc=1,
            msg=f"The container [ {self.container_name} ] failed to start. Check to lxc is available and that the container is in a functional state.",
        )

    def _check_archive(self):
        """Create a compressed archive of a container.

        This will store archive_info in as self.archive_info
        """

        if self.module.params["archive"]:
            self.archive_info = {"archive": self._container_create_tar()}

    def _check_clone(self):
        """Create a compressed archive of a container.

        This will store archive_info in as self.archive_info
        """

        clone_name = self.module.params["clone_name"]
        if clone_name:
            if not self._container_exists(container_name=clone_name, lxc_path=self.lxc_path):
                self.clone_info = {"cloned": self._container_create_clone()}
            else:
                self.clone_info = {"cloned": False}

    def _destroyed(self, timeout=60):
        """Ensure a container is destroyed.

        :param timeout: Time before the destroy operation is abandoned.
        :type timeout: ``int``
        """

        for dummy in range(timeout):
            if not self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
                break

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()

            if self._get_state() != "stopped":
                self.state_change = True
                self.container.stop()

            if self.container.destroy():
                self.state_change = True

            # post destroy attempt sleep for 1 second.
            time.sleep(1)
        else:
            self.failure(
                lxc_container=self._container_data(),
                error=f"Failed to destroy container [ {self.container_name} ]",
                rc=1,
                msg=(
                    f"The container [ {self.container_name} ] failed to be destroyed. "
                    "Check that lxc is available and that the container is in a functional state."
                ),
            )

    def _frozen(self, count=0):
        """Ensure a container is frozen.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method="frozen")
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            self._execute_command()

            # Perform any configuration updates
            self._config()

            container_state = self._get_state()
            if container_state == "frozen":
                pass
            elif container_state == "running":
                self.container.freeze()
                self.state_change = True
            else:
                self._container_startup()
                self.container.freeze()
                self.state_change = True

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._frozen(count)

    def _restarted(self, count=0):
        """Ensure a container is restarted.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method="restart")
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            self._execute_command()

            # Perform any configuration updates
            self._config()

            if self._get_state() != "stopped":
                self.container.stop()
                self.state_change = True

            # Run container startup
            self._container_startup()

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._restarted(count)

    def _stopped(self, count=0):
        """Ensure a container is stopped.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method="stop")
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            self._execute_command()

            # Perform any configuration updates
            self._config()

            if self._get_state() != "stopped":
                self.container.stop()
                self.state_change = True

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._stopped(count)

    def _started(self, count=0):
        """Ensure a container is started.

        If the container does not exist the container will be created.

        :param count: number of times this command has been called by itself.
        :type count: ``int``
        """

        self.check_count(count=count, method="start")
        if self._container_exists(container_name=self.container_name, lxc_path=self.lxc_path):
            container_state = self._get_state()
            if container_state == "running":
                pass
            elif container_state == "frozen":
                self._unfreeze()
            elif not self._container_startup():
                self.failure(
                    lxc_container=self._container_data(),
                    error=f"Failed to start container [ {self.container_name} ]",
                    rc=1,
                    msg=f"The container [ {self.container_name} ] failed to start. Check to lxc is available and that the container is in a functional state.",
                )

            # Return data
            self._execute_command()

            # Perform any configuration updates
            self._config()

            # Check if the container needs to have an archive created.
            self._check_archive()

            # Check if the container is to be cloned
            self._check_clone()
        else:
            self._create()
            count += 1
            self._started(count)

    def _get_lxc_vg(self):
        """Return the name of the Volume Group used in LXC."""

        build_command = [self.module.get_bin_path("lxc-config", True), "lxc.bdev.lvm.vg"]
        rc, vg, err = self.module.run_command(build_command)
        if rc != 0:
            self.failure(err=err, rc=rc, msg="Failed to read LVM VG from LXC config", command=" ".join(build_command))
        else:
            return str(vg.strip())

    def _lvm_lv_list(self):
        """Return a list of all lv in a current vg."""

        vg = self._get_lxc_vg()
        with lvs_runner(self.module)("noheadings separator fields vg") as ctx:
            rc, stdout, err = ctx.run(separator=";", fields="lv_name", vg=[vg])
        if self.module._verbosity >= 4:
            self.run_info.append(ctx.run_info)
        if rc != 0:
            self.failure(err=err, rc=rc, msg="Failed to get list of LVs")
        return [line.strip() for line in stdout.splitlines() if line.strip()]

    def _get_vg_free_pe(self, vg_name):
        """Return the available size of a given VG.

        :param vg_name: Name of volume.
        :type vg_name: ``str``
        :returns: size and measurement of an LV
        :type: ``tuple``
        """

        with vgs_runner(self.module)("noheadings nosuffix units separator fields vg") as ctx:
            rc, stdout, err = ctx.run(units="g", separator=";", fields="vg_free", vg=[vg_name])
        if self.module._verbosity >= 4:
            self.run_info.append(ctx.run_info)
        if rc != 0:
            self.failure(err=err, rc=rc, msg=f"failed to read vg {vg_name}")
        return float(stdout.strip()), "g"

    def _get_lv_size(self, lv_name):
        """Return the available size of a given LV.

        :param lv_name: Name of volume.
        :type lv_name: ``str``
        :returns: size and measurement of an LV
        :type: ``tuple``
        """

        vg = self._get_lxc_vg()
        lv = os.path.join(vg, lv_name)
        with lvs_runner(self.module)("noheadings nosuffix units separator fields vg") as ctx:
            rc, stdout, err = ctx.run(units="g", separator=";", fields="lv_size", vg=[lv])
        if self.module._verbosity >= 4:
            self.run_info.append(ctx.run_info)
        if rc != 0:
            self.failure(err=err, rc=rc, msg=f"failed to read lv {lv}")
        return self._roundup(float(stdout.strip())), "g"

    def _lvm_snapshot_create(self, source_lv, snapshot_name, snapshot_size_gb=5):
        """Create an LVM snapshot.

        :param source_lv: Name of lv to snapshot
        :type source_lv: ``str``
        :param snapshot_name: Name of lv snapshot
        :type snapshot_name: ``str``
        :param snapshot_size_gb: Size of snapshot to create
        :type snapshot_size_gb: ``int``
        """

        vg = self._get_lxc_vg()
        free_space, measurement = self._get_vg_free_pe(vg_name=vg)

        if free_space < float(snapshot_size_gb):
            message = (
                f"Snapshot size [ {snapshot_size_gb} ] is > greater than [ {free_space} ] on volume group [ {vg} ]"
            )
            self.failure(error="Not enough space to create snapshot", rc=2, msg=message)

        with lvcreate_runner(self.module)("size_L is_snapshot lv vg") as ctx:
            rc, dummy, err = ctx.run(
                size_L=f"{snapshot_size_gb}g",
                is_snapshot=True,
                lv=[snapshot_name],
                vg=[os.path.join(vg, source_lv)],
            )
        if self.module._verbosity >= 4:
            self.run_info.append(ctx.run_info)
        if rc != 0:
            self.failure(err=err, rc=rc, msg=f"Failed to Create LVM snapshot {vg}/{source_lv} --> {snapshot_name}")

    def _lvm_lv_mount(self, lv_name, mount_point):
        """mount an lv.

        :param lv_name: name of the logical volume to mount
        :type lv_name: ``str``
        :param mount_point: path on the file system that is mounted.
        :type mount_point: ``str``
        """

        vg = self._get_lxc_vg()

        build_command = [
            self.module.get_bin_path("mount", True),
            f"/dev/{vg}/{lv_name}",
            mount_point,
        ]
        rc, stdout, err = self.module.run_command(build_command)
        if rc != 0:
            self.failure(err=err, rc=rc, msg=f"failed to mountlvm lv {vg}/{lv_name} to {mount_point}")

    def _create_tar(self, source_dir):
        """Create an archive of a given ``source_dir`` to ``output_path``.

        :param source_dir:  Path to the directory to be archived.
        :type source_dir: ``str``
        """

        old_umask = os.umask(int("0077", 8))

        archive_path = self.module.params["archive_path"]
        if not os.path.isdir(archive_path):
            os.makedirs(archive_path)

        archive_compression = self.module.params["archive_compression"]
        compression_type = LXC_COMPRESSION_MAP[archive_compression]

        # remove trailing / if present.
        archive_name = f"{os.path.join(archive_path, self.container_name)}.{compression_type['extension']}"

        build_command = [
            self.module.get_bin_path("tar", True),
            f"--directory={os.path.realpath(source_dir)}",
            compression_type["argument"],
            archive_name,
            ".",
        ]

        rc, stdout, err = self.module.run_command(build_command)

        os.umask(old_umask)

        if rc != 0:
            self.failure(err=err, rc=rc, msg="failed to create tar archive", command=" ".join(build_command))

        return archive_name

    def _lvm_lv_remove(self, lv_name):
        """Remove an LV.

        :param lv_name: The name of the logical volume
        :type lv_name: ``str``
        """

        vg = self._get_lxc_vg()
        with lvremove_runner(self.module)("force lv") as ctx:
            rc, dummy, err = ctx.run(force=True, lv=[f"{vg}/{lv_name}"])
        if self.module._verbosity >= 4:
            self.run_info.append(ctx.run_info)
        if rc != 0:
            self.failure(err=err, rc=rc, msg=f"Failed to remove LVM LV {vg}/{lv_name}")

    def _rsync_data(self, container_path, temp_dir):
        """Sync the container directory to the temp directory.

        :param container_path: path to the container container
        :type container_path: ``str``
        :param temp_dir: path to the temporary local working directory
        :type temp_dir: ``str``
        """
        # This loop is created to support overlayfs archives. This should
        # squash all of the layers into a single archive.
        fs_paths = container_path.split(":")
        if "overlayfs" in fs_paths:
            fs_paths.pop(fs_paths.index("overlayfs"))

        for fs_path in fs_paths:
            # Set the path to the container data
            fs_path = os.path.dirname(fs_path)

            # Run the sync command
            build_command = [
                self.module.get_bin_path("rsync", True),
                "-aHAX",
                fs_path,
                temp_dir,
            ]
            rc, stdout, err = self.module.run_command(
                build_command,
            )
            if rc != 0:
                self.failure(err=err, rc=rc, msg="failed to perform archive", command=" ".join(build_command))

    def _unmount(self, mount_point):
        """Unmount a file system.

        :param mount_point: path on the file system that is mounted.
        :type mount_point: ``str``
        """

        build_command = [
            self.module.get_bin_path("umount", True),
            mount_point,
        ]
        rc, stdout, err = self.module.run_command(build_command)
        if rc != 0:
            self.failure(err=err, rc=rc, msg=f"failed to unmount [ {mount_point} ]", command=" ".join(build_command))

    def _overlayfs_mount(self, lowerdir, upperdir, mount_point):
        """mount an lv.

        :param lowerdir: name/path of the lower directory
        :type lowerdir: ``str``
        :param upperdir: name/path of the upper directory
        :type upperdir: ``str``
        :param mount_point: path on the file system that is mounted.
        :type mount_point: ``str``
        """

        build_command = [
            self.module.get_bin_path("mount", True),
            "-t",
            "overlayfs",
            "-o",
            f"lowerdir={lowerdir},upperdir={upperdir}",
            "overlayfs",
            mount_point,
        ]
        rc, stdout, err = self.module.run_command(build_command)
        if rc != 0:
            self.failure(
                err=err,
                rc=rc,
                msg=f"failed to mount overlayfs:{lowerdir}:{upperdir} to {mount_point} -- Command: {build_command}",
            )

    def _container_create_tar(self):
        """Create a tar archive from an LXC container.

        The process is as follows:
            * Stop or Freeze the container
            * Create temporary dir
            * Copy container and config to temporary directory
            * If LVM backed:
                * Create LVM snapshot of LV backing the container
                * Mount the snapshot to tmpdir/rootfs
            * Restore the state of the container
            * Create tar of tmpdir
            * Clean up
        """

        with tempfile.TemporaryDirectory() as temp_dir:
            # Set the name of the working dir, temp + container_name
            work_dir = os.path.join(temp_dir, self.container_name)

            # LXC container rootfs
            lxc_rootfs = self.container.get_config_item("lxc.rootfs")

            # Test if the containers rootfs is a block device
            block_backed = lxc_rootfs.startswith(os.path.join(os.sep, "dev"))

            # Test if the container is using overlayfs
            overlayfs_backed = lxc_rootfs.startswith("overlayfs")

            mount_point = os.path.join(work_dir, "rootfs")

            # Set the snapshot name if needed
            snapshot_name = f"{self.container_name}_lxc_snapshot"

            container_state = self._get_state()
            try:
                # Ensure the original container is stopped or frozen
                if container_state not in ["stopped", "frozen"]:
                    if container_state == "running":
                        self.container.freeze()
                    else:
                        self.container.stop()

                # Sync the container data from the container_path to work_dir
                self._rsync_data(lxc_rootfs, temp_dir)

                if block_backed:
                    if snapshot_name not in self._lvm_lv_list():
                        if not os.path.exists(mount_point):
                            os.makedirs(mount_point)

                        # Take snapshot
                        size, measurement = self._get_lv_size(lv_name=self.container_name)
                        self._lvm_snapshot_create(
                            source_lv=self.container_name, snapshot_name=snapshot_name, snapshot_size_gb=size
                        )

                        # Mount snapshot
                        self._lvm_lv_mount(lv_name=snapshot_name, mount_point=mount_point)
                    else:
                        self.failure(
                            err=f"snapshot [ {snapshot_name} ] already exists",
                            rc=1,
                            msg=f"The snapshot [ {snapshot_name} ] already exists. Please clean up old snapshot of containers before continuing.",
                        )
                elif overlayfs_backed:
                    lowerdir, upperdir = lxc_rootfs.split(":")[1:]
                    self._overlayfs_mount(lowerdir=lowerdir, upperdir=upperdir, mount_point=mount_point)

                # Set the state as changed and set a new fact
                self.state_change = True
                return self._create_tar(source_dir=work_dir)
            finally:
                if block_backed or overlayfs_backed:
                    # unmount snapshot
                    self._unmount(mount_point)

                if block_backed:
                    # Remove snapshot
                    self._lvm_lv_remove(snapshot_name)

                # Restore original state of container
                if container_state == "running":
                    if self._get_state() == "frozen":
                        self.container.unfreeze()
                    else:
                        self.container.start()

    def check_count(self, count, method):
        if count > 1:
            self.failure(
                error=f"Failed to {method} container",
                rc=1,
                msg=f"The container [ {self.container_name} ] failed to {method}. Check to lxc is available and that the container is in a functional state.",
            )

    def failure(self, **kwargs):
        """Return a Failure when running an Ansible command.

        :param error: ``str``  Error that occurred.
        :param rc: ``int``     Return code while executing an Ansible command.
        :param msg: ``str``    Message to report.
        """
        if self.run_info:
            kwargs["run_info"] = self.run_info

        self.module.fail_json(**kwargs)

    def run(self):
        """Run the main method."""

        action = getattr(self, LXC_ANSIBLE_STATES[self.state])
        action()

        outcome = self._container_data()
        if self.archive_info:
            outcome.update(self.archive_info)

        if self.clone_info:
            outcome.update(self.clone_info)

        self.module.exit_json(changed=self.state_change, lxc_container=outcome)


def main():
    """Ansible Main module."""

    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            template=dict(type="str", default="ubuntu"),
            backing_store=dict(type="str", choices=list(LXC_BACKING_STORE.keys()), default="dir"),
            template_options=dict(type="str"),
            config=dict(
                type="path",
            ),
            vg_name=dict(type="str", default="lxc"),
            thinpool=dict(type="str"),
            fs_type=dict(type="str", default="ext4"),
            fs_size=dict(type="str", default="5G"),
            directory=dict(type="path"),
            zfs_root=dict(type="str"),
            lv_name=dict(type="str"),
            lxc_path=dict(type="path"),
            state=dict(choices=list(LXC_ANSIBLE_STATES.keys()), default="started"),
            container_command=dict(type="str"),
            container_config=dict(type="list", elements="str"),
            container_log=dict(type="bool", default=False),
            container_log_level=dict(choices=[n for i in LXC_LOGGING_LEVELS.values() for n in i], default="INFO"),
            clone_name=dict(
                type="str",
            ),
            clone_snapshot=dict(type="bool", default="false"),
            archive=dict(type="bool", default=False),
            archive_path=dict(
                type="path",
            ),
            archive_compression=dict(choices=list(LXC_COMPRESSION_MAP.keys()), default="gzip"),
        ),
        supports_check_mode=False,
        required_if=([("archive", True, ["archive_path"])]),
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    if not HAS_LXC:
        module.fail_json(msg="The `lxc` module is not importable. Check the requirements.")

    if not module.params["lv_name"]:
        module.params["lv_name"] = module.params["name"]

    lxc_manage = LxcContainerManagement(module=module)
    lxc_manage.run()


if __name__ == "__main__":
    main()
